from __future__ import annotations

import dataclasses
import datetime

from typing import Any

from mknodes.basenodes import mktext
from mknodes.info import yamlfile
from mknodes.navs import mknav
from mknodes.pages import metadata, mkpage
from mknodes.utils import resources


@dataclasses.dataclass(frozen=True)
class Author:
    """A blog post author."""

    name: str
    description: str | None = None
    avatar: str | None = None


class BlogMetadata(metadata.Metadata):
    """Extended Metadata class for blog posts."""

    @property
    def date(self) -> str | None:
        return self.get("date")

    @date.setter
    def date(self, val: str | None):
        self["date"] = val

    @property
    def draft(self) -> bool | None:
        return self.get("draft")

    @draft.setter
    def draft(self, val: bool | None):
        self["draft"] = val

    @property
    def categories(self) -> list[str] | None:
        return self.get("categories")

    @categories.setter
    def categories(self, val: list[str] | None):
        self["categories"] = val

    @property
    def authors(self) -> list[str] | None:
        return self.get("authors")

    @authors.setter
    def authors(self, val: list[str] | None):
        self["authors"] = val


class MkBlog(mknav.MkNav):
    """Class representing the blog provided by the MkDocs-Material blog plugin."""

    def __init__(self, section: str | None = "Blog", **kwargs: Any):
        super().__init__(section=section, **kwargs)
        self.authors: dict[str, Author] = {}
        # self.posts = mknav.MkNav("posts", parent=self)
        # self.index_page = mkpage.MkPage()

    def add_author(
        self,
        username: str,
        name: str,
        description: str | None = None,
        avatar: str | None = None,
    ):
        """Add an author to the blog.

        Authors get linked to the blog post authors to show extended information.

        Arguments:
            username: Username / slug of the author
            name: Full name of the author
            description: Short description of the author
            avatar: Url to an image used as avatar
        """
        author = Author(name=name, description=description, avatar=avatar)
        self.authors[username] = author

    def get_node_resources(self) -> resources.Resources:
        authors_file = yamlfile.YamlFile()
        dct = {k: dataclasses.asdict(v) for k, v in self.authors.items()}
        authors_file._data["authors"] = dct
        content = authors_file.serialize("yaml")
        return resources.Resources(
            assets=[
                resources.Asset("blog/.authors.yml", content, target="docs_dir"),
            ],
        )

    def add_post(
        self,
        text: str,
        date: datetime.datetime | str,
        more_text: str | None = None,
        *,
        categories: list[str] | str | None = None,
        authors: list[str] | str | None = None,
        draft: bool = False,
        **kwargs: Any,
    ):
        """Add a post to the blog.

        Arguments:
            text: Blog post text
            date: Time/Date of the post
            more_text: Text hidden behint the more button.
            categories: A list of categories this post is linked to
            authors: A list of authors of the post
            draft: Whether post is a draft
            kwargs: Keyword arguments passed to parent
        """
        page = MkBlogPost(
            date=date,
            draft=draft,
            categories=categories,
            authors=authors,
            **kwargs,
        )
        page += mktext.MkText(text)
        if more_text:
            page += "\n\n<!-- more -->\n\n"
            page += more_text
        self += page


class MkBlogPost(mkpage.MkPage):
    """Class representing a MkDocs-Material blog post."""

    DATE_FORMAT = "%Y-%m-%d"

    def __init__(
        self,
        date: str | datetime.datetime,
        *,
        draft: bool = True,
        categories: list[str] | str | None = None,
        authors: list[str] | str | None = None,
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            date: Time/Date of the post
            draft: Whether post is a draft
            categories: A list of categories this post is linked to
            authors: A list of authors of the post
            kwargs: Keyword arguments passed to parent
        """
        super().__init__(**kwargs)
        if isinstance(authors, str):
            authors = [authors]
        if isinstance(categories, str):
            categories = [categories]
        self.metadata = BlogMetadata(
            **self.metadata,
            date=(
                date
                if isinstance(date, str)
                else datetime.datetime.strftime(date, self.DATE_FORMAT)
            ),
            draft=draft,
            categories=categories or [],
            authors=authors or [],
        )


if __name__ == "__main__":
    blog = MkBlog()
    blog.add_post("Hello", "2020-01-01", "more")
    page = blog.nav.pages[0]
    print(page.metadata)
    print(blog)



from __future__ import annotations

import os
import tomllib

from typing import Any, Literal
from xml.etree import ElementTree as Et

from mknodes.basenodes import mkcontainer, mknode
from mknodes.utils import log, pathhelpers, reprhelpers, resources, xmlhelpers as xml


logger = log.get_logger(__name__)

SCROLLREVEAL_LINK = (
    "https://cdn.jsdelivr.net/npm/scrollreveal@3.4.0/dist/scrollreveal.min.js"
)
JQUERY_LINK = "https://code.jquery.com/jquery-2.2.4.min.js"

STYLE = (
    "background: linear-gradient(rgba(0, 0, 0, 0), rgba(0, 0, 0,"
    " 0.4)), url({image}) center center no-repeat;"
    " background-size: cover;"
)

SCRIPT = """
window.sr = ScrollReveal();

// Add class to <html> if ScrollReveal is supported
// Note: only works in version 3
if (sr.isSupported()) {
document.documentElement.classList.add('sr');
}
"""


class MkTimelineItem(mknode.MkNode):
    """Single Timeline item / card.

    Not intended to be used directly. Use MkTimeline instead, which manages
    MkTimelineItems.
    """

    def __init__(
        self,
        title: str = "",
        content: str | mknode.MkNode | Et.Element = "",
        label: str = "",
        link: str = "",
        button_text: str = "More",
        image: str = "",
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            title: Item header / title
            content: Text / markdown for the content area
            label: A label, displayed in an upper corner
            link: An optional link for the "More" button
            button_text: allows to switch the button label from "More" to sth user-chosen.
            image: Optional image to display in upper half of the card
            kwargs: Keyword arguments passed to parent
        """
        super().__init__(**kwargs)
        self.title = title
        self.content = self.to_child_node(content)
        self.image = image
        self.label = label
        self.link = link
        self.button_text = button_text
        self.fade_direction: Literal["left", "right"] | None = None

    @property
    def children(self):
        return [self.content]

    @children.setter
    def children(self, val):
        pass

    def get_element(self) -> xml.Div:
        root = xml.Div("timeline-item")
        xml.Div("timeline-img", parent=root)
        match self.fade_direction:
            case "left":
                fade = " js--fadeInLeft"
            case "right":
                fade = " js--fadeInRight"
            case _:
                fade = ""
        tl = " timeline-card" if self.image else ""
        content_div = xml.Div(f"timeline-content{tl}{fade}", parent=root)
        if self.image:
            style = STYLE.format(image=self.image)
            header_div = xml.Div("timeline-img-header", parent=content_div, style=style)
            p = xml.P(parent=header_div)
            xml.Header(2, self.title, parent=p)
        elif self.title:
            xml.Header(2, self.title, parent=content_div)
        if self.label:
            xml.Div("date", text=self.label, parent=content_div)
        p_text = xml.P(parent=content_div)
        match self.content:
            case str():
                p_text.text = self.content
            case Et.Element():
                p_text.append(self.content)
            case _:
                text = f"<div>{self.content.to_html()}</div>"
                p_text.append(Et.fromstring(text))
        if self.link:
            xml.A("bnt-more", href=self.link, text=self.button_text, parent=content_div)
        return root


class MkTimeline(mkcontainer.MkContainer):
    """Node to show a JavaScript-supported Timeline.

    Consists of cards which slide in and out once they enter / leave the screen.
    """

    ICON = "material/timeline"
    JS_FILES = [
        resources.JSFile(JQUERY_LINK, is_library=True),
        resources.JSFile(SCROLLREVEAL_LINK, is_library=True),
        resources.JSText(SCRIPT, "scrollreveal.js"),  # type: ignore[list-item]
        resources.JSFile("js/timeline.js"),
    ]
    CSS = [resources.CSSFile("css/timeline.css")]
    items: list[MkTimelineItem]

    def __init__(
        self,
        items: list | str | os.PathLike | None = None,
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            items: Timeline items or a path to a TOML file containing timeline data
            kwargs: Keyword arguments passed to parent
        """
        if isinstance(items, str | os.PathLike):
            text = pathhelpers.load_file_cached(str(items))
            data = tomllib.loads(text)
            items = [MkTimelineItem(**step) for step in data.values()]
        super().__init__(items, **kwargs)

    def __repr__(self):
        return reprhelpers.get_repr(self)

    def get_element(self) -> xml.Section:
        root = xml.Section("timeline")
        div = xml.Div(parent=root)
        for i, item in enumerate(self.items):
            item.fade_direction = "left" if i % 2 == 0 else "right"
            elem = item.get_element()
            div.append(elem)
        return root

    def _to_markdown(self) -> str:
        root = self.get_element()
        return "\n\n" + root.to_string() + "\n\n"

    def add_item(
        self,
        title: str = "",
        content: str | mknode.MkNode = "",
        label: str = "",
        link: str = "",
        button_text: str = "More",
        image: str = "",
        **kwargs: Any,
    ):
        """Add a timeline item.

        title: Item header
        content: Markdown for content
        label: Label to be displayed in small box at the top
        link: Optional button-link. Text of button can be set via button_text
        button_text: Text for the link button.
        image: Optional image for the item
        kwargs: keyword arguments passed to parent
        """
        item = MkTimelineItem(
            title=title,
            content=content,
            label=label,
            link=link,
            button_text=button_text,
            image=image,
            **kwargs,
        )
        self += item
        return item

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        node = MkTimeline()
        for i in range(1, 6):
            node += MkTimelineItem(
                title=f"Image card {i}",
                content="A card with an image.",
                label=f"{i} JANUARY 2023",
                link="https://phil65.github.io/mknodes",
                image=f"https://picsum.photos/40{i}",
            )
            admonition = mk.MkAdmonition("test")
            node += MkTimelineItem(
                title=f"Card {i}",
                content=admonition,
                label=f"{i} JANUARY 2023",
                link="https://phil65.github.io/mknodes",
            )
        page += node


if __name__ == "__main__":
    import mknodes as mk

    node = mk.MkAdmonition("test")
    item = MkTimelineItem(
        title="Title",
        content=node,
        label="1 MAY 2016",
        link="https://phil65.github.io/mknodes",
        # image="https://picsum.photos/1000/800/?random",
    )
    item2 = MkTimelineItem(
        title="Title",
        content="fdsfs",
        label="1 MAY 2016",
        link="https://phil65.github.io/mknodes",
    )
    timeline = MkTimeline.with_default_context([item, item2, item])



from __future__ import annotations

import mknodes as mk


def test_virtual_files():
    nav = mk.MkNav()
    subnav = nav.add_nav("subsection")
    page = subnav.add_page("page")
    img = mk.MkBinaryImage(data=b"", path="Test.jpg")
    page.append(img)



from __future__ import annotations

from typing import Any

from mknodes.basenodes import mkcontainer
from mknodes.utils import log, reprhelpers


logger = log.get_logger(__name__)


class MkJinjaTemplate(mkcontainer.MkContainer):
    """Node representing a jinja template.

    Renders templates with the context-aware MkNodes jinja environment.
    Rendered nodes become virtual children of this node.
    Additional variables can be passed to the render process.
    """

    ICON = "simple/jinja"
    STATUS = "new"

    def __init__(
        self,
        template: str,
        *,
        variables: dict[str, Any] | None = None,
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            template: Jinja template name.
            variables: Variables to use for rendering
            kwargs: Keyword arguments passed to parent
        """
        super().__init__(**kwargs)
        self.template = template
        self.variables = variables or {}

    def __repr__(self):
        return reprhelpers.get_repr(
            self,
            template=self.template,
            variables=self.variables,
            _filter_empty=True,
        )

    @property
    def items(self):
        self.env.render_template(self.template, variables=self.variables)
        return self.env.rendered_nodes

    @items.setter
    def items(self, val):
        pass

    @classmethod
    def create_example_page(cls, page):
        node = MkJinjaTemplate(template="nodes_index.jinja")
        page += node

    def _to_markdown(self) -> str:
        return self.env.render_template(self.template, variables=self.variables)


if __name__ == "__main__":
    node = MkJinjaTemplate("nodes_index.jinja")
    print(node.get_resources())



from __future__ import annotations

import dataclasses

from mknodes.utils import helpers, log


logger = log.get_logger(__name__)

# // Primary color shades
# --md-primary-fg-color:               hsla(#{hex2hsl($clr-indigo-500)}, 1);
# --md-primary-fg-color--light:        hsla(#{hex2hsl($clr-indigo-400)}, 1);
# --md-primary-fg-color--dark:         hsla(#{hex2hsl($clr-indigo-700)}, 1);
# --md-primary-bg-color:               hsla(0, 0%, 100%, 1);
# --md-primary-bg-color--light:        hsla(0, 0%, 100%, 0.7);

# // Accent color shades
# --md-accent-fg-color:                hsla(#{hex2hsl($clr-indigo-a200)}, 1);
# --md-accent-fg-color--transparent:   hsla(#{hex2hsl($clr-indigo-a200)}, 0.1);
# --md-accent-bg-color:                hsla(0, 0%, 100%, 1);
# --md-accent-bg-color--light:         hsla(0, 0%, 100%, 0.7);


@dataclasses.dataclass
class ColorTheme:
    color: str
    light_shade: str | None = None
    dark_shade: str | None = None

    @property
    def color_str(self) -> str:
        return helpers.get_color_str(self.color)

    @property
    def light_str(self) -> str:
        return helpers.get_color_str(self.light_shade or self.color)

    @property
    def dark_str(self) -> str:
        return helpers.get_color_str(self.dark_shade or self.color)


if __name__ == "__main__":
    color_theme = ColorTheme("#FFFFFF")
    print(color_theme)



from __future__ import annotations

import os
import pathlib

from typing import Any, Self

from mknodes.basenodes import mkimage
from mknodes.utils import helpers, icons, log


logger = log.get_logger(__name__)


class MkBinaryImage(mkimage.MkImage):
    """Image carrying the data by itself.

    This node basically is a regular image link, but carries the image data by itself.
    The data will get written to the "virtual" folder at the end of the process.
    It can hold either str or bytes as data.
    """

    ICON = "material/file-image"

    def __init__(self, data: bytes | str, path: str, **kwargs: Any):
        """Constructor.

        Arguments:
            data: Image data
            path: path for the image (including extension)
            kwargs: Keyword arguments passed to parent
        """
        super().__init__(path=path, **kwargs)
        self.data = data

    @property
    def files(self) -> dict[str, str | bytes]:
        path = "/".join(self.resolved_parts) + "/" + self.path
        return {path: self.data} | self._files

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        page += "A BinaryImage carries the image data by itself."
        page += "A file containing the data will become part of the file tree later on."
        data = """<svg width="90pt" height="90pt" version="1.1"
        viewBox="0 0 15.875 10.583" xmlns="http://www.w3.org/2000/svg">
         <g fill="none" stroke="#F00" stroke-width=".3">
          <path d="m6.1295 3.6601 3.2632 3.2632z"/>
          <path d="m9.3927 3.6601-3.2632 3.2632z"/>
         </g>
        </svg>
        """
        node = MkBinaryImage(data, path="some_image.svg", caption="A simple cross")
        page += mk.MkHeader("From data", level=3)
        page += mk.MkReprRawRendered(node)
        node = MkBinaryImage.for_icon("file-image", width=200)
        page += mk.MkHeader("From icon", level=3)
        page += mk.MkReprRawRendered(node)

    @classmethod
    def for_icon(cls, icon: str, **kwargs: Any) -> Self:
        """Return a MkBinaryImage with data for given icon.

        Arguments:
            icon: Icon to get a MkBinaryImage for (example: material/file-image)
            kwargs: Keyword arguments passed to constructor
        """
        content = icons.get_icon_svg(icon)
        path = f"{helpers.slugify(icon)}.svg"
        return cls(data=content, path=path, **kwargs)

    @classmethod
    def for_file(cls, path: str | os.PathLike, **kwargs: Any) -> Self:
        """Return a MkBinaryImage with data for given icon.

        Arguments:
            path: Icon to get a MkBinaryImage for (example: material/file-image)
            kwargs: Keyword arguments passed to constructor
        """
        path = pathlib.Path(path)
        return cls(data=path.read_bytes(), path=path.name, **kwargs)


if __name__ == "__main__":
    node = MkBinaryImage.for_file("assets/cli.gif")
    print(node.data)



from __future__ import annotations

from mknodes.basenodes import mknode
from mknodes.utils import log, reprhelpers


logger = log.get_logger(__name__)

HTML = '<iframe frameborder="0" width="{width}" height="{height}" src="{url}"></iframe>'


class MkIFrame(mknode.MkNode):
    """Node for embedding an IFrame."""

    ICON = "material/web-box"

    def __init__(
        self,
        url: str,
        *,
        width: int = 300,
        height: int = 150,
        header: str = "",
    ):
        """Constructor.

        Arguments:
            url: Url to display in a frame
            width: width of frame in pixels
            height: height of frame in pixels
            header: Section header
        """
        # if as_links:
        #     url = [link.Link(i) for i in url]
        super().__init__(header=header)
        self.url = url
        self.width = width
        self.height = height

    def __repr__(self):
        return reprhelpers.get_repr(
            self,
            url=self.url,
            width=self.width,
            height=self.height,
        )

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        frame = MkIFrame(url="https://phil65.github.io/mknodes/", width=600, height=600)
        page += mk.MkReprRawRendered(frame)

    def _to_markdown(self) -> str:
        if not self.url:
            return ""
        return HTML.format(url=self.url, width=self.width, height=self.height)


if __name__ == "__main__":
    url = "http://www.google.de"
    # "https://jex.im/regulex/#!embed=true&flags=&re=%5E(a%7Cb)*%3F%24"
    section = MkIFrame(url, header="test")
    print(section)



from __future__ import annotations

import textwrap

from typing import TYPE_CHECKING, Any

from mknodes.basenodes import mkcontainer
from mknodes.utils import log


if TYPE_CHECKING:
    import mknodes as mk

logger = log.get_logger(__name__)


class MkBlock(mkcontainer.MkContainer):
    """PyMdown-based block."""

    ICON = "material/cube"

    def __init__(
        self,
        name: str,
        content: list | str | mk.MkNode = "",
        *,
        argument: str = "",
        attributes: dict[str, Any] | None = None,
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            name: Block name
            content: Block content
            argument: Block argument
            attributes: Block attributes
            kwargs: Keyword arguments passed to parent
        """
        super().__init__(content=content, **kwargs)
        self.name = name
        self.attributes = attributes or {}
        self.argument = argument

    @property
    def fence_boundary(self) -> str:
        """Return the fence boundary ("///") based on nesting depth."""
        block_level = sum(isinstance(i, MkBlock) for i in self.ancestors)
        return "/" * (block_level + 3)

    @property
    def content_block(self) -> str:
        """Returns the block content. Can be reimplemented by subclasses."""
        return textwrap.indent(super()._to_markdown(), self.indent).rstrip("\n") + "\n"

    @property
    def attributes_block(self) -> str:
        """The text block for attributes.

        For subclasses, implement a property for self.attributes.
        """
        if not self.attributes:
            return ""
        lines = [f"    {k}: {v}" for k, v in self.attributes.items() if v is not None]
        return "\n".join(lines) + "\n"

    def _to_markdown(self) -> str:
        boundary = self.fence_boundary
        base = f"{boundary} {self.name}"
        if self.argument:
            base += f" | {self.argument}"
        return f"{base}\n{self.attributes_block}\n{self.content_block}{boundary}\n"

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        page += "An MkBlock is a base class for pymdownx-style blocks."
        block = MkBlock("html", content="Some text", argument="div")
        page += mk.MkCode(str(block), header="### Markdown")
        # We can also nest blocks, they will adjust their delimiters automatically.
        page += mk.MkHtmlBlock(block, header="### Nested")


if __name__ == "__main__":
    inner_1 = MkBlock("inner_1", content="inner_1 content")
    inner_2 = MkBlock("inner_2", content="inner_2 content")
    outer = MkBlock("outer", content=[inner_1, inner_2])
    print(outer)



from __future__ import annotations

import os
import pathlib

from typing import Any

from mknodes.info import grifferegistry
from mknodes.pages import mktemplatepage
from mknodes.utils import classhelpers, log, reprhelpers


logger = log.get_logger(__name__)


DEFAULT_TPL = "classpage.md"


class MkClassPage(mktemplatepage.MkTemplatePage):
    """Page showing information about a class."""

    def __init__(
        self,
        klass: type,
        *,
        path: str | os.PathLike | None = None,
        module_path: tuple[str, ...] | str | None = None,
        template: str | os.PathLike | None = None,
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            klass: class to show info for
            module_path: If given, overrides module returned by class.__module__
                         This can be useful if you want to link to an aliased class
                         (for example a class imported to __init__.py)
            template: Name of the template to load
            path: Filename/path for the class page. defaults to [classname].md
            kwargs: keyword arguments passed to base class
        """
        self.klass = klass
        match module_path:
            case None:
                self.parts = klass.__module__.split(".")
            case _:
                self.parts = classhelpers.to_module_parts(module_path)
        # if user chooses custom template, we make default the parent
        tpl = template or DEFAULT_TPL
        super().__init__(
            path=path or pathlib.Path(f"{klass.__name__}.md"),
            template=tpl,
            template_parent=DEFAULT_TPL if tpl != DEFAULT_TPL else None,
            **kwargs,
        )

    def __repr__(self):
        return reprhelpers.get_repr(self, klass=self.klass, path=self.path)

    @property
    def extra_variables(self) -> dict[str, Any]:
        # right now, we inject the cls and the griffe Class into jinja namespace.
        subclasses = list(classhelpers.iter_subclasses(self.klass, recursive=False))
        variables = dict(cls=self.klass, subclasses=subclasses)
        variables["griffe_obj"] = grifferegistry.registry.get_class(self.klass)
        return variables


if __name__ == "__main__":
    import mknodes as mk

    doc = MkClassPage.with_default_context(mk.Environment)
    print(doc.to_markdown())



from __future__ import annotations

import mknodes as mk


EXPECTED = """## Header

[^1]:
    abcde
    fghi
[^2]:
    abcde
    fghi
[^3]:
    abcde
    fghi
[^4]:
    abcde
    fghi
[^5]:
    abcde
    fghi
[^6]:
    abcde
    fghi
[^7]:
    abcde
    fghi
[^8]:
    abcde
    fghi
[^9]:
    abcde
    fghi
[^10]:
    abcde
    fghi
"""

EXPECTED_SORTED = """[^1]:
    1
[^2]:
    2
"""


def test_empty():
    annotation = mk.MkFootNotes()
    assert not str(annotation)


def test_if_annotations_get_sorted():
    node = mk.MkFootNotes()
    node[2] = "2"
    node[1] = "1"
    assert str(node) == EXPECTED_SORTED


def test_markdown():
    annotation = mk.MkFootNotes(["abcde\nfghi"] * 10, header="Header")
    assert str(annotation) == EXPECTED


def test_constructors():
    annotation_1 = mk.MkFootNotes(["abc", "def"])
    anns = {1: "abc", 2: "def"}
    annotation_2 = mk.MkFootNotes(anns)
    assert str(annotation_1) == str(annotation_2)


def test_mapping_interface():
    ann = mk.MkFootNotes()
    ann[1] = "test"
    assert str(ann[1]) == "[^1]:\n    test\n"



from __future__ import annotations

from typing import Any

from mknodes.basenodes import mknode
from mknodes.utils import log, reprhelpers, resources, xmlhelpers as xml


logger = log.get_logger(__name__)

# https://ruyadorno.com/simple-slider/

JS_URL = "https://rawgit.com/ruyadorno/simple-slider/master/dist/simpleslider.min.js"

SCRIPT = """\
window.addEventListener('DOMContentLoaded', function () {
  simpleslider.getSlider();
})
"""


class MkImageSlideshow(mknode.MkNode):
    """Node to show an Image slideshow (in autoplay mode)."""

    ICON = "material/image-multiple"
    JS_FILES: list[resources.JSFile | resources.JSText] = [
        resources.JSFile(JS_URL),
        resources.JSText(SCRIPT, "slideshow.js"),
    ]

    def __init__(
        self,
        images: list[str],
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            images: A list of image URLs to use for the slideshow
            kwargs: Keyword arguments passed to parent
        """
        super().__init__(**kwargs)
        self.images = images

    def __repr__(self):
        return reprhelpers.get_repr(
            self,
            images=self.images,
            _filter_empty=True,
            _filter_false=True,
        )

    def get_element(self) -> xml.Div:
        attrs = {"data-simple-slider": ""}
        root = xml.Div(style="width: 100%; padding-bottom: 25%", parent=None, **attrs)
        for img in self.images:
            xml.Img(src=img, parent=root)
        return root

    def _to_markdown(self) -> str:
        root = self.get_element()
        return root.to_string()

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        images = ["https://picsum.photos/700", "https://picsum.photos/701"]
        node = MkImageSlideshow(images)
        page += mk.MkReprRawRendered(node)


if __name__ == "__main__":
    img = MkImageSlideshow(["https://picsum.photos/200", "https://picsum.photos/201"])



from __future__ import annotations

import dataclasses

from mknodes.nodemods import modhelpers
from mknodes.utils import helpers, resources


SCRIPT = """\
var image = document.getElementsByClassName('%s');
new simpleParallax(image, %s);
"""

LIB = "https://cdn.jsdelivr.net/npm/simple-parallax-js@5.5.1/dist/simpleParallax.min.js"


@dataclasses.dataclass(frozen=True)
class ParallaxEffect:
    orientation: str = "up"
    scale: float = 1.5
    overflow: bool = False
    delay: float = 0.6
    transition: str = "cubic-bezier(0,0,0,1)"
    # max_transition: int = 100

    def get_resources(self):
        file = resources.JSFile(LIB, is_library=True)
        dct = dataclasses.asdict(self)
        js_map = modhelpers.format_js_map(dct)
        script = resources.JSText(
            SCRIPT % (self.css_class_names[0], js_map),
            filename="parallax.js",
        )
        return resources.Resources(js=[file, script])

    @property
    def css_class_names(self):
        return [f"parallax_{helpers.get_hash(self)}"]


if __name__ == "__main__":
    effect = ParallaxEffect(orientation="down")
    print(effect.css_class_names)



from __future__ import annotations

import os

from typing import Any

from mknodes.basenodes import mknode
from mknodes.pages import mkpage
from mknodes.utils import log, pathhelpers, reprhelpers


EXAMPLE_URL = "https://raw.githubusercontent.com/phil65/mknodes/main/README.md"


logger = log.get_logger(__name__)


class MkInclude(mknode.MkNode):
    """Node to include the text of other Markdown files / MkNodes.

    This node only keeps a reference to given target and resolves it when needed.
    Target can either be an URL or any other fsspec protocol path to a markdown file,
    a file system path to a markdown file, or another MkNode.
    """

    ICON = "octicons/file-symlink-file-24"

    def __init__(
        self,
        target: str | os.PathLike | mknode.MkNode,
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            target: target to include
            kwargs: Keyword arguments passed to parent
        """
        super().__init__(**kwargs)
        self.target = target

    def __repr__(self):
        return reprhelpers.get_repr(self, target=self.target)

    def _to_markdown(self) -> str:  # type: ignore[return]
        match self.target:
            case os.PathLike() | str():
                return pathhelpers.load_file_cached(str(self.target))
            case mknode.MkNode():
                return str(self.target)
            case _:
                raise TypeError(self.target)

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        node = MkInclude(EXAMPLE_URL)
        page += mk.MkReprRawRendered(node)


if __name__ == "__main__":
    page = mkpage.MkPage(content="test")
    include = MkInclude(target=EXAMPLE_URL)
    print(include)



from __future__ import annotations

import mknodes as mk


EXPECTED_IN_ANNOTATIONS = """1.  ::: mknodes.basenodes._mkdocstrings.MkDocStrings.__init__
        options:
          docstring_section_style: 'list'
          show_root_heading: True
    """ + "\n"  # noqa: ISC003


def test_docstrings():
    docstrings = mk.MkDocStrings(obj=mk)
    assert str(docstrings) == "::: mknodes\n"


def test_auto_list_style_inside_annotations():
    annotations = mk.MkAnnotations()
    annotations[1] = mk.MkDocStrings(obj=mk.MkDocStrings.__init__)
    assert str(annotations) == EXPECTED_IN_ANNOTATIONS



from __future__ import annotations

import textwrap

from typing import Any

from mknodes.basenodes import mkblock, mkcontainer, mknode
from mknodes.utils import log, reprhelpers, resources


logger = log.get_logger(__name__)


class MkTabBlock(mkblock.MkBlock):
    """Node representing a single tab (new block style)."""

    ICON = "material/tab"
    REQUIRED_EXTENSIONS = [resources.Extension("pymdownx.blocks.tab")]

    def __init__(
        self,
        title: str,
        content: str | mknode.MkNode | list,
        *,
        new: bool | None = None,
        select: bool | None = None,
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            title: Tab title
            content: Tab content
            new: Whether tab should start a new tab bloock
            select: Whether tab should be initially selected
            kwargs: Keyword arguments passed to parent
        """
        super().__init__(
            "tab",
            content=content,
            argument=title,
            **kwargs,
        )
        if new is not None:
            self.new = new
        if select is not None:
            self.select = select

    @property
    def title(self):
        return self.argument

    @title.setter
    def title(self, value):
        self.argument = value

    @property
    def new(self):
        return self.attributes.get("new", False)

    @new.setter
    def new(self, value: bool):
        self.attributes["new"] = value

    @property
    def select(self):
        return self.attributes.get("select", False)

    @select.setter
    def select(self, value: bool):
        self.attributes["select"] = value


class MkTab(mkcontainer.MkContainer):
    """Node representing a single tab."""

    ICON = "material/tab"
    REQUIRED_EXTENSIONS = [resources.Extension("pymdownx.tabbed")]
    ATTR_LIST_SEPARATOR = "    "

    def __init__(
        self,
        title: str,
        content: list | str | mknode.MkNode | None = None,
        *,
        new: bool = False,
        select: bool = False,
        attrs: dict | None = None,
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            title: Tab title
            content: Tab content
            new: Whether tab should start a new tab bloock
            select: Whether tab should be initially selected
            attrs: Additional attributes for the tab
            kwargs: Keyword arguments passed to parent
        """
        self.title = title
        self.select = select
        self.new = new
        self.attrs = attrs
        super().__init__(content=content, **kwargs)

    def __repr__(self):
        if len(self.items) == 1:
            content = reprhelpers.to_str_if_textnode(self.items[0])
        else:
            content = [reprhelpers.to_str_if_textnode(i) for i in self.items]
        return reprhelpers.get_repr(
            self,
            title=self.title,
            content=content,
            select=self.select,
            new=self.new,
            attrs=self.attrs,
            _filter_empty=True,
            _filter_false=True,
        )

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        # We can add single tabs to a page by themselves.
        # It is recommended to use a Tab container though.
        tab = MkTab("A Title", content="Tab content(1)")
        tab.annotations[1] = "Tabs can carry annotations."
        page += mk.MkReprRawRendered(tab, header="### With annotations")

    def attach_annotations(self, text: str) -> str:
        # we deal with attaching annotations ourselves.
        return text

    def _to_markdown(self) -> str:
        text = "\n\n".join(i.to_markdown() for i in self.items)
        text = text.rstrip("\n")
        if self.annotations:
            annotates = str(self.annotations)
            text = f"{text}\n{{ .annotate }}\n\n{annotates}"
        text = textwrap.indent(text, prefix="    ")
        if self.new:
            mark = "!"
        elif self.select:
            mark = "+"
        else:
            mark = ""
        lines = [f'==={mark} "{self.title}"', text]
        return "\n".join(lines) + "\n"


if __name__ == "__main__":
    tab = MkTabBlock(content="test", title="test", new=True)
    print(tab)



from __future__ import annotations

from collections.abc import Sequence
import inspect
import types

from mknodes.basenodes import mktable
from mknodes.utils import classhelpers, helpers, layouts, log, reprhelpers


logger = log.get_logger(__name__)


class MkModuleTable(mktable.MkTable):
    """Node for a table containing formatted information about a module."""

    ICON = "material/view-module-outline"

    def __init__(
        self,
        modules: Sequence[types.ModuleType | str | tuple[str, ...]],
        **kwargs,
    ):
        self.modules = [classhelpers.to_module(i, return_none=False) for i in modules]
        super().__init__(**kwargs)

    def __repr__(self):
        return reprhelpers.get_repr(self, modules=self.modules)

    @property
    def data(self):
        if not self.modules:
            return {}
        layout = layouts.ModuleLayout(link_provider=self.ctx.links)
        data = [layout.get_row_for(mod) for mod in self.modules]
        return {
            k: [self.to_child_node(dic[k]) for dic in data]  # type: ignore[index]
            for k in data[0]
        }

    @classmethod
    def create_example_page(cls, page):
        import jinja2

        import mknodes as mk

        modules = [mod for _, mod in inspect.getmembers(jinja2, inspect.ismodule)]
        node = MkModuleTable(modules=modules)
        page += mk.MkReprRawRendered(node)


if __name__ == "__main__":
    table = MkModuleTable(modules=[mktable, helpers, classhelpers])
    print(table)



from __future__ import annotations

from mknodes.pages import metadata


EXPECTED = """---
description: Some description
hide:
- toc
- path
icon: material/emoticon-happy
search:
  boost: 2.0
  exclude: false
status: new
subtitle: Some subtitle
title: Some title
---
"""

example_metadata = metadata.Metadata(
    hide=["toc", "path"],
    search_boost=2.0,
    exclude_from_search=False,
    icon="material/emoticon-happy",
    status="new",
    title="Some title",
    subtitle="Some subtitle",
    description="Some description",
)


def test_metadata():
    assert example_metadata.as_page_header() == EXPECTED


def test_no_string_if_empty():
    data = metadata.Metadata()
    assert not data.as_page_header()


def test_parsing_metadata():
    parsed, _rest = metadata.Metadata.parse(EXPECTED)
    assert parsed.as_page_header() == EXPECTED



from __future__ import annotations

import mknodes as mk


def test_equality():
    node_1 = mk.MkHeader("test")
    str(node_1)  # generate build stats
    node_2 = mk.MkHeader("test")
    str(node_2)  # generate build stats
    assert node_1 == node_2



from __future__ import annotations

from typing import Any

from mknodes.basenodes import mknode
from mknodes.utils import clihelpers, log, reprhelpers


logger = log.get_logger(__name__)


class MkClickDoc(mknode.MkNode):
    """Node for showing documentation for click / typer CLI apps."""

    ICON = "material/api"

    def __init__(
        self,
        target: str | None = None,
        prog_name: str | None = None,
        show_hidden: bool = False,
        show_subcommands: bool = False,
        **kwargs: Any,
    ):
        r"""Constructor.

        Arguments:
            target: Dotted path to Click command
            prog_name: Program name
            show_hidden: Show commands and options that are marked as hidden.
            show_subcommands: List subcommands of a given command.
            kwargs: Keyword arguments passed to parent
        """
        super().__init__(**kwargs)
        self.target = target
        self.prog_name = prog_name
        self.show_hidden = show_hidden
        self.show_subcommands = show_subcommands

    def __repr__(self):
        return reprhelpers.get_repr(
            self,
            target=self.target,
            prog_name=self.prog_name,
            show_hidden=self.show_hidden,
            show_subcommands=self.show_subcommands,
            _filter_false=True,
            _filter_empty=True,
        )

    @property
    def attributes(self) -> dict[str, Any]:
        # sourcery skip: use-named-expression
        dct: dict[str, Any] = {}
        match self.target:
            case str():
                module, command = self.target.split(":")
                dct = dict(module=module, command=command, prog_name=self.prog_name)
            case None:
                if cli_eps := self.ctx.metadata.entry_points.get("console_scripts"):
                    module, command = cli_eps[0].dotted_path.split(":")
                    dct = dict(module=module, command=command, prog_name=cli_eps[0].name)
        if not dct:
            return {}
        dct.update(show_hidden=self.show_hidden, show_subcommands=self.show_subcommands)
        return dct

    def _to_markdown(self) -> str:
        import importlib

        if not self.attributes:
            return ""
        attrs = self.attributes
        mod = importlib.import_module(attrs["module"])
        instance = getattr(mod, attrs["command"])
        info = clihelpers.get_typer_info(instance, command=attrs["prog_name"])
        return info.to_markdown(recursive=self.show_subcommands)

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        node = MkClickDoc(target="mkdocs_mknodes.cli:cli")
        page += mk.MkReprRawRendered(node)
        node = MkClickDoc(target="mkdocs_mknodes.cli:cli", prog_name="build")
        page += mk.MkReprRawRendered(node)


if __name__ == "__main__":
    docstrings = MkClickDoc.with_default_context("mkdocs_mknodes.cli:cli")
    print(docstrings)



from __future__ import annotations

import os
import pathlib
import re

import fsspec

from mknodes.utils import log


logger = log.get_logger(__name__)

_RFC_3986_PATTERN = re.compile(r"^[A-Za-z][A-Za-z0-9+\-+.]*://")


def is_fsspec_url(url: str | os.PathLike[str]) -> bool:
    """Returns true if the given URL looks like something fsspec can handle."""
    return (
        isinstance(url, str)
        and bool(_RFC_3986_PATTERN.match(url))
        and not url.startswith(("http://", "https://"))
    )


class FsSpecPath:
    def __init__(
        self,
        path: dict | str,
        fs: fsspec.AbstractFileSystem | str,
        **kwargs,
    ):
        filesystem = fsspec.filesystem(fs, **kwargs) if isinstance(fs, str) else fs
        match path:
            case dict():
                self.path = path
            case str() if filesystem:
                self.path = filesystem.info(path)
            case FsSpecPath():
                self.path = path.path
        self.fs = filesystem

    def __str__(self):
        return self.protocol_path

    def __repr__(self):
        return f"FsSpecPath({self.protocol_path!r})"

    def __truediv__(self, other):
        path = str(pathlib.Path(self.path["name"]) / other)
        return FsSpecPath(path, self.fs)

    @property
    def parent(self):
        path = str(pathlib.Path(self.path["name"]).parent)
        return FsSpecPath(path, self.fs)

    @property
    def name(self):
        return pathlib.Path(str(path)).name if (path := self.path["name"]) else ""

    @property
    def protocol_path(self) -> str:
        """Get protocol path for given index."""
        protocol = self.fs.protocol
        path = self.path["name"]
        return f"{protocol}://{path}"

    def iterdir(self):
        try:
            for i in self.fs.ls(self.path["name"], detail=True):
                yield FsSpecPath(i, self.fs)
        except Exception as e:  # noqa: BLE001
            logger.warning(
                "Error fetching %s for protocol %s: %s",
                self.path["name"],
                self.fs.protocol,
                e,
            )
            return

    def read_text(self) -> str:
        with self.fs.open(self.protocol_path) as file:
            content = file.read()
            return content.decode() if isinstance(content, bytes) else content

    def is_file(self) -> bool:
        return self.path["type"] == "file"

    def is_dir(self) -> bool:
        return self.path["type"] == "directory"

    def exists(self) -> bool:
        return self.fs.exists(self.path["name"])

    def absolute(self):
        return self


if __name__ == "__main__":
    path = FsSpecPath("", "github", org="phil65", repo="mknodes")

    # from mknodes.treelib import filetreenode
    # folder = filetreenode.FileTreeNode.from_folder(
    #     path,
    #     exclude_folders=["__pycache__", ".git", ".mypy_cache"],
    #     sort=False,
    #     maximum_depth=2,
    # )
    # logger.warning(folder.get_tree_repr())



from __future__ import annotations

from mknodes.utils import connector, log


logger = log.get_logger(__name__)


class NodeConnector(connector.Connector):
    def get_children(self, item):
        return item.children

    def get_id(self, item):
        # id() would be enough, but name is sometimes useful for debugging.
        return f"{type(item).__name__}_{id(item)}"

    def get_title(self, item) -> str:
        return f"{type(item).__name__}"


def to_tree_graph(node, direction: str = "TD") -> str:
    """Returns markdown to display a tree graph of this node and all subnodes.

    Arguments:
        node: node to get a graph for
        direction: Direction of resulting graph
    """
    item_str = NodeConnector([node]).get_graph_connection_text()
    text = f"graph {direction}\n{item_str}"
    return f"```mermaid\n{text}\n```"



from __future__ import annotations

import dataclasses

from typing import ClassVar, Literal

from mknodes.utils import log


logger = log.get_logger(__name__)

InstallMethodStr = Literal["pip", "pipx", "conda_forge", "homebrew"]


@dataclasses.dataclass
class InstallMethod:
    ID: ClassVar
    project: str

    def info_text(self) -> str:
        raise NotImplementedError

    def install_instructions(self) -> str:
        raise NotImplementedError

    @classmethod
    def get_installmethods(cls):
        return {i.ID: i for i in cls.__subclasses__()}

    @classmethod
    def by_id(cls, identifier: str):
        return cls.get_installmethods()[identifier]


class PipInstall(InstallMethod):
    ID = "pip"

    def info_text(self):
        return (
            "The latest released version is available at the [Python "
            f"package index](https://pypi.org/project/{self.project})."
        )

    def install_instructions(self):
        return f"pip install {self.project}"


class PipXInstall(InstallMethod):
    ID = "pipx"

    def info_text(self):
        return (
            "[pipx](https://github.com/pypa/pipx) allows for the "
            "global installation of Python applications in isolated environments."
        )

    def install_instructions(self):
        return f"pipx install {self.project}"


class CondaForgeInstall(InstallMethod):
    ID = "conda_forge"

    def info_text(self):
        return (
            "See the "
            f"[feedstock](https://github.com/conda-forge/{self.project}-feedstock) "
            "for more details."
        )

    def install_instructions(self):
        return f"conda install -c conda-forge {self.project}"


class HomebrewInstall(InstallMethod):
    ID = "homebrew"

    def info_text(self):
        return (
            f"See the [formula](https://formulae.brew.sh/formula/{self.project}) "
            "for more details."
        )

    def install_instructions(self):
        return f"brew install {self.project}"


if __name__ == "__main__":
    instructions = PipInstall("mknodes")
    print(instructions)



from __future__ import annotations

from typing import Any

from mknodes.basenodes import mknode
from mknodes.utils import log, reprhelpers


logger = log.get_logger(__name__)


class MkHeader(mknode.MkNode):
    """Super simple header node."""

    ICON = "material/format-header-pound"
    STATUS = "new"

    def __init__(
        self,
        text: str | mknode.MkNode | None = "",
        level: int = 2,
        exclude_from_search: bool = False,
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            text: Header text
            level: Header level
            exclude_from_search: Whether section should be included in search.
            kwargs: Keyword arguments passed to parent
        """
        super().__init__(**kwargs)
        self.text = str(text or "")
        self.level = level
        self._exclude_from_search = exclude_from_search

    def __repr__(self):
        return reprhelpers.get_repr(self, text=self.text, level=self.level)

    def _to_markdown(self) -> str:
        level_str = "#" * self.level
        suffix = " { data-search-exclude }" if self._exclude_from_search else ""
        return f"{level_str} {self.text}{suffix}"

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        for i in range(1, 7):
            node = MkHeader(f"Level {i}", level=i)
            page += mk.MkReprRawRendered(node)


if __name__ == "__main__":
    header = MkHeader("Header!", level=1)
    print(header)



from __future__ import annotations

from mknodes.utils import downloadhelpers, log


logger = log.get_logger(__name__)

URL = "https://pypi.org/pypi/{name}/json"


class PyPiInfo:
    def __init__(self, pkg_name: str):
        self.package_name = pkg_name
        url = URL.format(name=self.package_name)
        self.response = downloadhelpers.download(url).decode()


if __name__ == "__main__":
    info = PyPiInfo("git_changelog")
    print(info.response)



from __future__ import annotations

from collections.abc import Sequence

import jinja2

from mknodes.info import yamlfile


class MkDocsConfigFile(yamlfile.YamlFile):
    @property
    def theme(self):
        """Return theme section."""
        return self.get_section("theme")

    @property
    def markdown_extensions(self) -> list:
        """Return a list of markdown extensions."""
        return self.get_section("markdown_extensions") or []

    @property
    def plugins(self) -> list:
        """Return a list of plugins from the config."""
        return self.get_section("plugins") or []

    @property
    def mknodes_config(self) -> dict:
        """Return our very own config section."""
        return self.get_section("plugins", "mknodes") or {}

    @property
    def mkdocstrings_config(self) -> dict:
        """Return the MkDocStrings config section."""
        return self.get_section("plugins", "mkdocstrings", "handlers", "python") or {}

    def update_mknodes_section(
        self,
        repo_url: str | None = None,
        build_fn: str | None = None,
        clone_depth: int | None = None,
    ):
        """Quick access to overriding our plugin settings.

        Arguments:
            repo_url: Repo url to set
            build_fn: Build function to set
            clone_depth: Amount of commits to fetch when cloning a repository.
        """
        for plugin in self._data["plugins"]:
            if "mknodes" in plugin:
                if repo_url is not None:
                    plugin["mknodes"]["repo_path"] = repo_url
                if build_fn is not None:
                    plugin["mknodes"]["build_fn"] = build_fn
                if clone_depth is not None:
                    plugin["mknodes"]["clone_depth"] = clone_depth

    def get_loaders(self) -> Sequence[jinja2.BaseLoader]:
        from mknodes.jinja import loaders

        jinja_loaders: list[jinja2.BaseLoader] = [
            loaders.FileSystemLoader(searchpath=self._data.get("docs_dir", "docs")),
        ]
        for loader_dct in self.mknodes_config.get("loaders", []):
            dct = loader_dct.copy()
            kls = loaders.LOADERS[dct.pop("type")]
            loader = kls(**dct)
            jinja_loaders.append(loader)
        return jinja_loaders

    def get_inventory_infos(self) -> list[dict]:
        """Returns list of dicts containing inventory info.

        Links are taken from mkdocstrings section.

        Shape: [{"url": inventory_url, "domains": ["std", "py"]}, ...]
        """
        return self.mkdocstrings_config.get("import") or []

    @property
    def theme_name(self) -> str:
        """Name of currently used theme."""
        try:
            return self._data["theme"]["name"]
        except (KeyError, TypeError):
            return self._data["theme"]


if __name__ == "__main__":
    info = MkDocsConfigFile("mkdocs.yml")
    print(info.get_loaders())



from __future__ import annotations

from collections.abc import MutableMapping
from typing import Any

from mknodes.data import admonition, datatypes
from mknodes.info import contexts
from mknodes.jinja import environment
from mknodes.pages import templateblocks, templateregistry
from mknodes.utils import helpers, icons, log, reprhelpers, resources


logger = log.get_logger(__name__)


class Theme:
    """MkDocs Theme."""

    css_template = "css/theme_mkdocs.css"

    def __init__(
        self,
        name: str,
        *,
        data: dict[str, Any] | None = None,
        template_registry: templateregistry.TemplateRegistry | None = None,
    ):
        self.name = name
        self.data = data or {}
        self.features = self.data.get("features")
        self.env = environment.Environment(load_templates=True)
        self.templates = template_registry or templateregistry.TemplateRegistry()
        self.main_template = self.templates["main.html"]
        self.error_page = self.templates["404.html"]

        self.alternating_table_colors = False

        self.admonitions: list[admonition.AdmonitionType] = []
        self.add_admonition_type(
            name="theme",
            icon="mdi:file",
            header_color=self.primary_color,
            icon_color="black",
            border_color="black",
            font_color=self.text_color,
        )

    def __repr__(self):
        return reprhelpers.get_repr(self, name=self.name)

    def add_admonition_type(
        self,
        name: str,
        icon: str,
        header_color: datatypes.ColorType,
        icon_color: datatypes.ColorType | None = None,
        border_color: datatypes.ColorType | None = None,
        font_color: datatypes.ColorType | None = None,
    ):
        """Add a custom admonition type.

        Arguments:
            name: Slug for new admonition type
            icon: Either a Pyconify icon key or an <svg> element
            header_color: Color for the admonition header
            icon_color: Color for the admonition icon
            border_color: Color for the admonition border
            font_color: Color for the admonition header font
        """
        header_col_str = helpers.get_color_str(header_color)
        icon_col_str = helpers.get_color_str(icon_color or (255, 255, 255))
        border_col_str = helpers.get_color_str(border_color or (255, 255, 255))
        font_col_str = helpers.get_color_str(font_color or (255, 255, 255))
        adm = admonition.AdmonitionType(
            name=name,
            svg=icons.get_icon_svg(icon) if not icon.startswith("<") else icon,
            header_color=header_col_str,
            icon_color=icon_col_str,
            border_color=border_col_str,
            font_color=font_col_str,
        )
        self.admonitions.append(adm)

    def get_resources(self) -> resources.Resources:
        """Return resources required for the theme.

        Usually, the resources consist of static templates and CSS.
        """
        req: list[resources.CSSFile | resources.CSSText] = []
        if self.css_template:
            tmpl_ctx = self.get_css_context()
            css_text = self.env.render_template(
                self.css_template,
                variables=tmpl_ctx,
            )
            req = [resources.CSSText(content=css_text, filename="mknodes_theme.css")]
        return resources.Resources(css=req)

    def get_css_context(self) -> dict[str, Any]:
        """Return variables used to resolve the CSS template.

        Can be overridden by subclasses.
        """
        return dict(
            admonitions=self.admonitions,
            primary_color=self.primary_color,
            text_color=self.text_color,
            alternating_table_colors=self.alternating_table_colors,
            css_primary_fg=self.text_color,
            css_primary_bg=self.primary_color,
            css_primary_bg_light=self.primary_color,
            css_accent_fg=self.text_color,
            css_accent_fg_transparent=self.text_color,
            css_accent_bg=self.primary_color,
            css_default_fg="#222222",
            css_default_bg="#DDDDDD",
        )

    @property
    def context(self):
        ctx = self.get_css_context()
        return contexts.ThemeContext(
            name=self.name,
            data=self.data,
            primary_color=self.primary_color,
            text_color=self.text_color,
            admonitions=self.admonitions,
            css_primary_fg=ctx["css_primary_fg"],
            css_primary_bg=ctx["css_primary_bg"],
            css_primary_bg_light=ctx["css_primary_bg_light"],
            css_accent_fg=ctx["css_accent_fg"],
            css_accent_fg_transparent=ctx["css_accent_fg_transparent"],
            css_accent_bg=ctx["css_accent_bg"],
            css_default_fg=ctx["css_default_fg"],
            css_default_bg=ctx["css_default_bg"],
        )

    @classmethod
    def get_theme(
        cls,
        theme_name: str = "material",
        data: dict[str, Any] | None = None,
        **kwargs: Any,
    ):
        """Return theme instance of corresponding subclass based on theme name.

        Arguments:
            theme_name: Name of the theme
            data: Additional data for the theme
            kwargs: Additional keyword arguments passe to the Theme
        """
        if theme_name == "material":
            import mknodes as mk

            return mk.MaterialTheme(data=data, **kwargs)
        return Theme(theme_name, data=data, **kwargs)

    def iter_nodes(self):
        import mknodes as mk

        for block in self.main_template.blocks:
            if not isinstance(block, templateblocks.HtmlBlock):
                continue
            for node in block.items:
                if not isinstance(node, mk.MkNode):
                    continue
                yield 0, node

    @property
    def primary_color(self) -> str:
        """Return the primary color of the theme.

        Can be overridden by subclasses.
        """
        return "#2fa4e7"

    @property
    def text_color(self) -> str:
        """Return the text color of the theme.

        Can be overridden by subclasses.
        """
        return "#333333"

    def adapt_extensions(self, extensions: MutableMapping[str, dict]):
        if self.name.lower() != "mkdocs":
            return
        for k in dict(extensions).copy():
            ext = extensions[k]
            if k == "pymdownx.highlight":
                # default style "table" looks broken with mkdocs
                ext["pymdownx.highlight"]["linenums_style"] = "inline"  # pymdownx-inline

    def adapt_extras(self, extras: dict):
        pass

    @property
    def template_path(self) -> str:
        from mkdocs import utils

        return utils.get_theme_dir(self.name)


if __name__ == "__main__":
    theme = Theme.get_theme("material")
    print(theme)



from __future__ import annotations

import dataclasses

from mknodes.utils import icons, xmlhelpers as xml


SCRIPT = """
  var palette = __md_get("__palette")
  if (palette && typeof palette.color === "object")
    for (var key of Object.keys(palette.color))
      document.body.setAttribute("data-md-color-" + key, palette.color[key])
"""


@dataclasses.dataclass
class Palette:
    """Class representing a MkDocs-Material palette."""

    scheme: str = "default"
    primary: str = "indigo"
    accent: str = "indigo"
    media: str | None = None
    toggle_name: str | None = None
    toggle_icon: str | None = None


def build_toggle(palettes: list[Palette]) -> xml.Form:
    """Builds the XML equivalent to the content of "palette.html" partial."""
    # body = ElementTree.Element("body")
    # form = ElementTree.SubElement(
    # body,
    kwargs = {"data-md-component": "palette"}
    form = xml.Form("md-header__option", parent=None, attrs=kwargs)
    for i, pal in enumerate(palettes):
        mode = "dark" if pal.scheme == "slate" else "light"
        attrs = {
            "class": "md-option",
            "data-md-color-media": f"(prefers-color-scheme: {mode})",
            "data-md-color-scheme": pal.scheme.replace(" ", "-"),
            "data-md-color-primary": pal.primary.replace(" ", "-"),
            "data-md-color-accent": pal.accent.replace(" ", "-"),
            "type": "radio",
            "name": "__palette",
            "id": f"__palette_{i + 1}",
        }
        if pal.toggle_name:
            attrs["aria-label"] = pal.toggle_name
        else:
            attrs["aria-hidden"] = "true"
        el_input = xml.Input(parent=form, attrs=attrs)
        if pal.toggle_name:
            attrs = {
                "class": "md-header__button md-icon",
                "title": pal.toggle_name,
                "for": f"__palette_{i or len(palettes)}",
                "hidden": "hidden",
            }
            label = xml.Label(parent=el_input, attrs=attrs)
            el = icons.get_icon_xml(pal.toggle_icon or "brightness-4")
            label.append(el)
    # script = ElementTree.SubElement(body, "script")
    # script.text = SCRIPT
    return form


if __name__ == "__main__":
    default_palette = Palette()
    dark_palette = Palette(
        toggle_name="Switch to light mode",
        toggle_icon="material/brightness-4",
        scheme="slate",
    )
    root = build_toggle([default_palette, dark_palette])
    xml_string = xml.pformat(root)
    print(xml_string)



from __future__ import annotations

import textwrap

from typing import Any, Literal, get_args

from mknodes.basenodes import mkcontainer, mknode
from mknodes.data import datatypes
from mknodes.utils import log, reprhelpers, resources


logger = log.get_logger(__name__)


class MkAdmonition(mkcontainer.MkContainer):
    """Admonition info box."""

    ICON = "octicons/info-16"
    ATTR_LIST_SEPARATOR = "    "
    REQUIRED_EXTENSIONS = [
        resources.Extension("admonition"),
        resources.Extension("pymdownx.details"),
        resources.Extension("pymdownx.superfences"),
    ]

    def __init__(
        self,
        content: str | list | mknode.MkNode,
        *,
        typ: datatypes.AdmonitionTypeStr | str = "info",
        title: str | None = None,
        collapsible: bool = False,
        expanded: bool = False,
        inline: Literal["left", "right"] | None = None,
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            content: Admonition content
            typ: Admonition type
            title: Optional Admonition title
            collapsible: Whether Admontion can get collapsed by user
            expanded: Initial state if collapsible is set
            inline: Whether admonition should rendered as inline block
            kwargs: Keyword arguments passed to parent
        """
        self.typ = typ
        self.title = title
        self.collapsible = collapsible
        self.inline = inline
        self.expanded = expanded
        super().__init__(content=content, **kwargs)

    def __repr__(self):
        if len(self.items) == 1:
            content = reprhelpers.to_str_if_textnode(self.items[0])
        else:
            content = [reprhelpers.to_str_if_textnode(i) for i in self.items]
        return reprhelpers.get_repr(
            self,
            content=content,
            typ=self.typ if self.typ != "info" else None,
            title=self.title,
            collapsible=self.collapsible,
            expanded=self.expanded,
            _filter_empty=True,
            _filter_false=True,
        )

    def attach_annotations(self, text: str) -> str:
        # we deal with attaching annotations ourselves.
        return text

    def _to_markdown(self) -> str:
        if not self.items and not self.title:
            return ""
        block_start = "???" if self.collapsible else "!!!"
        if self.collapsible and self.expanded:
            block_start += "+"
        if self.inline:
            inline_label = " inline" if self.inline == "left" else " inline end"
        else:
            inline_label = ""
        if self.annotations:
            ann_marker = " annotate"
            annotations = f"\n{self.annotations}\n"
        else:
            ann_marker = ""
            annotations = ""
        title = f' "{self.title}"' if self.title is not None else ""
        text = "\n".join(i.to_markdown() for i in self.items)
        indented = textwrap.indent(text, "    ")
        optional = ann_marker + inline_label
        return f"{block_start} {self.typ}{optional}{title}\n{indented}\n{annotations}"

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        node = mk.MkAdmonition("MkAdmonitions can carry annotations(1).")
        node.annotations[1] = "Super handy!"
        page += mk.MkReprRawRendered(node)
        # AdmonitionTypeStr is a Literal containing all Admonition types
        for typ in get_args(datatypes.AdmonitionTypeStr):
            page += mk.MkHeader(f"Type '{typ}'", level=3)
            content = f"This is type **{typ}**"
            node = mk.MkAdmonition(typ=typ, content=content)
            page += mk.MkReprRawRendered(node)
        page += mk.MkHeader("Collapsible and expandable", level=3)
        node = mk.MkAdmonition(
            content="Admonitions can also be collapsible.",
            collapsible=True,
            title="Expand me!",
            expanded=True,  # this changes the initial state to expanded
        )
        page += mk.MkReprRawRendered(node)
        page += mk.MkHeader("Inlined", level=3)
        node = mk.MkAdmonition(content="Inlined.", inline="left", title="Inlined.")
        page += mk.MkReprRawRendered(node)


if __name__ == "__main__":
    admonition = MkAdmonition("")
    print(admonition)



from __future__ import annotations

import collections
import contextlib
import json
import pathlib

from typing import Any, Literal

import yaml

from mknodes.utils import log


logger = log.get_logger(__name__)
YAMLError = yaml.YAMLError

LoaderStr = Literal["unsafe", "full", "safe"]


def _patch_dumper_to_not_order(dumper_cls):
    def map_representer(dumper_cls_, data):
        return dumper_cls_.represent_dict(data.items())

    yaml.add_representer(dict, map_representer, Dumper=dumper_cls)
    yaml.add_representer(collections.OrderedDict, map_representer, Dumper=dumper_cls)


def patch_pyyaml_to_not_order_dicts():
    _dumpers = [getattr(yaml.dumper, x) for x in yaml.dumper.__all__]
    with contextlib.suppress(AttributeError):
        _cyaml = yaml.cyaml.__all__
        _dumpers += [getattr(yaml.cyaml, x) for x in _cyaml if x.endswith("Dumper")]
    for dumper in _dumpers:
        _patch_dumper_to_not_order(dumper)


# patch_pyyaml_to_not_order_dicts()


def yaml_include_constructor(loader: yaml.BaseLoader, node: yaml.Node) -> Any:
    """Include file referenced with !include node."""
    scalar = loader.construct_scalar(node)  # type: ignore[arg-type]
    folder = pathlib.Path(loader.name).parent
    fp = folder.joinpath(scalar).resolve()  # type: ignore[arg-type]
    fe = fp.suffix.lstrip(".")
    with fp.open() as f:
        if fe in ("yaml", "yml"):
            return yaml.load(f, type(loader))
        if fe in ("json", "jsn"):
            return json.load(f)
        return f.read()


def get_safe_loader(base_loader_cls: type):
    class SafeLoader(base_loader_cls):
        """Safe Loader."""

    SafeLoader.add_constructor("!relative", lambda loader, node: None)  # type: ignore
    SafeLoader.add_multi_constructor(
        "tag:yaml.org,2002:python/name:",
        lambda loader, suffix, node: None,
    )
    SafeLoader.add_multi_constructor(
        "tag:yaml.org,2002:python/object/apply:",
        lambda loader, suffix, node: None,
    )
    return SafeLoader


def get_default_loader(base_loader_cls: type):
    # Attach Environment Variable constructor.
    # See https://github.com/waylan/pyyaml-env-tag

    import yaml_env_tag

    class DefaultLoader(base_loader_cls):
        """Default Loader."""

    DefaultLoader.add_constructor("!ENV", yaml_env_tag.construct_env_tag)
    DefaultLoader.add_constructor("!include", yaml_include_constructor)
    # Loader.add_constructor("!ENV", lambda loader, node: None)  # type: ignore
    # if config is not None:
    #     Loader.add_constructor(
    #         "!relative", functools.partial(_construct_dir_placeholder, config)
    #     )
    return DefaultLoader


def load_yaml(text: str, mode: LoaderStr = "unsafe"):
    """Load a yaml string.

    Arguments:
        text: the string to load
        mode: the yaml loader mode.
    """
    match mode:
        case "unsafe":
            base_loader_cls: type = yaml.CUnsafeLoader
        case "full":
            base_loader_cls = yaml.CFullLoader
        case _:
            base_loader_cls = yaml.CSafeLoader

    # Derive from global loader to leave the global loader unaltered.
    default_loader = get_default_loader(base_loader_cls)
    return yaml.load(text, Loader=default_loader)


def dump_yaml(yaml_obj: Any) -> str:
    """Dump a data structure to a yaml string.

    Arguments:
        yaml_obj: The object to serialize
    """
    return yaml.dump(yaml_obj, Dumper=yaml.Dumper, indent=2)


if __name__ == "__main__":
    cfg = load_yaml("- test")
    print(cfg)



from __future__ import annotations

import dataclasses

from typing import Literal


# ![Static Badge](https://img.shields.io/badge/built_with-mknodes-yellow?link=www.pypa.org)


@dataclasses.dataclass
class Badge:
    identifier: str
    title: str
    image_url: str
    url: str
    group: str = ""

    def get_image_url(self, user, project, branch):
        return self.image_url.format(user=user, project=project, branch=branch)

    def get_url(self, user, project):
        return self.url.format(user=user, project=project)


# PYPI

latest_version_badge = Badge(
    identifier="version",
    group="pypi",
    title="PyPI Latest Version",
    image_url="https://img.shields.io/pypi/v/{project}.svg",
    url="https://pypi.org/project/{project}/",
)

license_badge = Badge(
    identifier="license",
    group="pypi",
    title="PyPI License",
    image_url="https://img.shields.io/pypi/l/{project}.svg",
    url="https://pypi.org/project/{project}/",
)

package_status_badge = Badge(
    identifier="status",
    group="pypi",
    title="Package status",
    image_url="https://img.shields.io/pypi/status/{project}.svg",
    url="https://pypi.org/project/{project}/",
)

weekly_downloads_badge = Badge(
    identifier="weekly_downloads",
    group="pypi",
    title="Weekly downloads",
    image_url="https://img.shields.io/pypi/dw/{project}.svg",
    url="https://pypi.org/project/{project}/",
)

daily_downloads_badge = Badge(
    identifier="daily_downloads",
    group="pypi",
    title="Daily downloads",
    image_url="https://img.shields.io/pypi/dd/{project}.svg",
    url="https://pypi.org/project/{project}/",
)

monthly_downloads_badge = Badge(
    identifier="monthly_downloads",
    group="pypi",
    title="Monthly downloads",
    image_url="https://img.shields.io/pypi/dm/{project}.svg",
    url="https://pypi.org/project/{project}/",
)


format_badge = Badge(
    identifier="format",
    group="pypi",
    title="Distribution format",
    image_url="https://img.shields.io/pypi/format/{project}.svg",
    url="https://pypi.org/project/{project}/",
)

wheel_badge = Badge(
    identifier="wheel",
    group="pypi",
    title="Wheel availability",
    image_url="https://img.shields.io/pypi/wheel/{project}.svg",
    url="https://pypi.org/project/{project}/",
)

python_version_badge = Badge(
    identifier="python_version",
    group="pypi",
    title="Python version",
    image_url="https://img.shields.io/pypi/pyversions/{project}.svg",
    url="https://pypi.org/project/{project}/",
)

implementation_badge = Badge(
    identifier="implementation",
    group="pypi",
    title="Implementation",
    image_url="https://img.shields.io/pypi/implementation/{project}.svg",
    url="https://pypi.org/project/{project}/",
)


# GITHUB

github_releases_badge = Badge(
    identifier="releases",
    group="github",
    title="Releases",
    image_url="https://img.shields.io/github/downloads/{user}/{project}/total.svg",
    url="https://github.com/{user}/{project}/releases",
)


github_contributors_badge = Badge(
    identifier="contributors",
    group="github",
    title="Github Contributors",
    image_url="https://img.shields.io/github/contributors/{user}/{project}",
    url="https://github.com/{user}/{project}/graphs/contributors",
)

github_discussions_badge = Badge(
    identifier="discussions",
    group="github",
    title="Github Discussions",
    image_url="https://img.shields.io/github/discussions/{user}/{project}",
    url="https://github.com/{user}/{project}/discussions",
)

github_forks_badge = Badge(
    identifier="forks",
    group="github",
    title="Github Forks",
    image_url="https://img.shields.io/github/forks/{user}/{project}",
    url="https://github.com/{user}/{project}/forks",
)

github_issues_badge = Badge(
    identifier="issues",
    group="github",
    title="Github Issues",
    image_url="https://img.shields.io/github/issues/{user}/{project}",
    url="https://github.com/{user}/{project}/issues",
)

github_pull_requests_badge = Badge(
    identifier="pull_requests",
    group="github",
    title="Github Issues",
    image_url="https://img.shields.io/github/issues-pr/{user}/{project}",
    url="https://github.com/{user}/{project}/pulls",
)

github_watchers_badge = Badge(
    identifier="watchers",
    group="github",
    title="Github Watchers",
    image_url="https://img.shields.io/github/watchers/{user}/{project}",
    url="https://github.com/{user}/{project}/watchers",
)


github_stars_badge = Badge(
    identifier="stars",
    group="github",
    title="Github Stars",
    image_url="https://img.shields.io/github/stars/{user}/{project}",
    url="https://github.com/{user}/{project}/stars",
)


github_repo_size_badge = Badge(
    identifier="repo_size",
    group="github",
    title="Github Repository size",
    image_url="https://img.shields.io/github/repo-size/{user}/{project}",
    url="https://github.com/{user}/{project}",
)

github_last_commit_badge = Badge(
    identifier="last_commit",
    group="github",
    title="Github last commit",
    image_url="https://img.shields.io/github/last-commit/{user}/{project}",
    url="https://github.com/{user}/{project}/commits",
)

github_release_date_badge = Badge(
    identifier="release_date",
    group="github",
    title="Github release date",
    image_url="https://img.shields.io/github/release-date/{user}/{project}",
    url="https://github.com/{user}/{project}/releases",
)


github_language_count_badge = Badge(
    identifier="language_count",
    group="github",
    title="Github language count",
    image_url="https://img.shields.io/github/languages/count/{user}/{project}",
    url="https://github.com/{user}/{project}",
)

# needs branch
# github_build_badge = Badge(
#     identifier="build",
#     group="github",
#     title="Github Build",
#     image_url="https://github.com/{user}/{project}/workflows/Build/badge.svg",
#     url="https://github.com/{user}/{project}/actions/",
# )

github_weekly_commits_badge = Badge(
    identifier="weekly_commits",
    group="github",
    title="Github commits this week",
    image_url="https://img.shields.io/github/commit-activity/w/{user}/{project}",
    url="https://github.com/{user}/{project}",
)

github_monthly_commits_badge = Badge(
    identifier="monthly_commits",
    group="github",
    title="Github commits this month",
    image_url="https://img.shields.io/github/commit-activity/m/{user}/{project}",
    url="https://github.com/{user}/{project}",
)

github_yearly_commits_badge = Badge(
    identifier="yearly_commits",
    group="github",
    title="Github commits this year",
    image_url="https://img.shields.io/github/commit-activity/y/{user}/{project}",
    url="https://github.com/{user}/{project}",
)


# MISC
code_cov_badge = Badge(
    identifier="codecov",
    group="quality",
    title="Package status",
    image_url="https://codecov.io/gh/{user}/{project}/branch/{branch}/graph/badge.svg",
    url="https://codecov.io/gh/{user}/{project}/",
)

black_badge = Badge(
    identifier="black",
    group="quality",
    title="Code style: black",
    image_url=r"https://img.shields.io/badge/code%20style-black-000000.svg",
    url="https://github.com/psf/black",
)

pyup_badge = Badge(
    identifier="pyup",
    group="dependencies",
    title="PyUp",
    image_url="https://pyup.io/repos/github/{user}/{project}/shield.svg",
    url="https://pyup.io/repos/github/{user}/{project}/",
)

SHIELDS = [
    latest_version_badge,
    license_badge,
    package_status_badge,
    daily_downloads_badge,
    weekly_downloads_badge,
    monthly_downloads_badge,
    format_badge,
    wheel_badge,
    python_version_badge,
    implementation_badge,
    # github_build_badge,
    github_releases_badge,
    github_contributors_badge,
    github_discussions_badge,
    github_forks_badge,
    github_issues_badge,
    github_pull_requests_badge,
    github_watchers_badge,
    github_stars_badge,
    github_repo_size_badge,
    github_last_commit_badge,
    github_release_date_badge,
    github_language_count_badge,
    github_weekly_commits_badge,
    github_monthly_commits_badge,
    github_yearly_commits_badge,
    code_cov_badge,
    black_badge,
    pyup_badge,
]

BadgeTypeStr = Literal[
    "version",
    "license",
    "status",
    "weekly_downloads",
    "daily_downloads",
    "monthly_downloads",
    "format",
    "wheel",
    "python_version",
    "implementation",
    "releases",
    "contributors",
    "discussions",
    "forks",
    "issues",
    "pull_requests",
    "watchers",
    "stars",
    "repo_size",
    "last_commit",
    "release_date",
    "language_count",
    "build",
    "weekly_commits",
    "monthly_commits",
    "yearly_commits",
    "codecov",
    "black",
    "pyup",
]



from __future__ import annotations

import dataclasses

from typing import Literal


RAW_GITHUB = "https://raw.githubusercontent.com"

BuildSystemStr = Literal[
    "hatch",
    "flit",
    "poetry",
    "setuptools",
    "pdm",
    "meson_python",
    "maturin",
]


@dataclasses.dataclass
class BuildSystem:
    identifier: BuildSystemStr
    build_backend: str
    url: str
    env_setup_cmd: str | None
    run_prefix: str = ""
    install_url: str | None = None


hatch = BuildSystem(
    identifier="hatch",
    build_backend="hatchling.build",
    url="https://hatch.pypa.io",
    env_setup_cmd="hatch env create",
    run_prefix="hatch run ",
    install_url=f"{RAW_GITHUB}/pypa/hatch/master/docs/install.md",
)
flit = BuildSystem(
    identifier="flit",
    build_backend="flit_core.buildapi",
    url="https://flit.pypa.io",
    env_setup_cmd="flit install",
)
poetry = BuildSystem(
    identifier="poetry",
    build_backend="poetry.core.masonry.api",
    url="https://python-poetry.org",
    env_setup_cmd="poetry install",
    run_prefix="poetry run ",
)
setuptools = BuildSystem(
    identifier="setuptools",
    build_backend="setuptools.build_meta",
    url="https://pypi.org/project/setuptools/",
    env_setup_cmd=None,
)
pdm = BuildSystem(
    identifier="pdm",
    build_backend="pdm.backend",
    url="https://pdm.fming.dev/",
    env_setup_cmd="pdm install",
    run_prefix="pdm run ",
    install_url=f"{RAW_GITHUB}/pdm-project/pdm/main/docs/docs/index.md#Installation",
)
mesonpy = BuildSystem(
    identifier="meson_python",
    build_backend="mesonpy",
    url="https://github.com/mesonbuild/meson-python",
    env_setup_cmd=None,
    install_url=(
        "https://meson-python.readthedocs.io/en/latest/how-to-guides/first-project.html"
    ),
)

maturin = BuildSystem(
    identifier="maturin",
    build_backend="maturin",
    url="https://www.maturin.rs",
    env_setup_cmd=None,
    install_url="https://www.maturin.rs/installation",
)

BUILD_SYSTEMS: dict[BuildSystemStr, BuildSystem] = {
    p.identifier: p for p in [hatch, flit, poetry, setuptools, pdm, mesonpy, maturin]
}



from __future__ import annotations

import re


CLASS_REGEX = re.compile(r"\b((mknodes\.Mk[A-Za-z\.\_]*).*)")


class AnnotationAppender:
    """Helper for auto-annotating code blocks.

    Example:
        test = "mknodes.MkAdmonition("test")"
        appender = AnnotationAppender()
        text = appender.append_markers(test)
        appender.append_annotations(node)

    """

    def __init__(self):
        self.count = 0
        self.matches = []

    def __call__(self, match) -> str:
        line = match.group()
        self.count += 1
        self.matches.append((self.count, match.group(2)))
        return f"{line} # ({self.count})"

    def append_markers(self, code: str) -> str:
        return re.sub(CLASS_REGEX, self, code) if "# (" not in code else code

    def append_annotations(self, node):
        import mknodes as mk

        for index, dotted_path in self.matches:
            node[index] = mk.MkDocStrings(dotted_path)



import mknodes as mk

from mknodes.manual import (
    dev_section,
    get_started_section,
    navs_section,
    nodes_section,
    page_section,
    templating_section,
)


def build(project: mk.Project[mk.MaterialTheme]) -> mk.MkNav:
    build = Build()
    project.linkprovider.add_inv_file("https://mkdocstrings.github.io/objects.inv")
    build.on_theme(project.theme)
    return build.on_root(project.root) or project.root


class Build:
    def on_theme(self, theme: mk.Theme):
        theme.error_page.content = mk.MkAdmonition("Page does not exist!")
        if isinstance(theme, mk.MaterialTheme):
            theme.content_area_width = 1300
            theme.tooltip_width = 800
            theme.add_status_icon("js", "fa6-brands:js", "Uses JavaScript")
            theme.add_status_icon("css", "vaadin:css", "Uses CSS")

    def on_root(self, nav: mk.MkNav):
        nav.page_template.announcement_bar = mk.MkMetadataBadges("websites")
        nav += get_started_section.nav
        nav += navs_section.nav
        nav += page_section.nav
        nav += nodes_section.nav
        nav += templating_section.nav
        nav += dev_section.nav
        return nav



from __future__ import annotations

from typing import Any

import mknodes as mk

from mknodes.utils import log


logger = log.get_logger(__name__)


class MkDefaultWebsite(mk.MkNav):
    """Nav for showing a default website including API docs and environment infos."""

    def __init__(
        self,
        static_pages: dict[str, str | dict | list] | None = None,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)

        page = self.add_page(is_index=True, hide="toc")
        page += mk.MkText(page.ctx.metadata.description)
        static_pages = static_pages or {}
        self.parse.json(static_pages)
        self.add_doc(section_name="API", recursive=True)
        self.page_template.announcement_bar = mk.MkMetadataBadges("websites")
        if self.ctx.metadata.cli:
            page = self.add_page("CLI", hide="nav")
            page += mk.MkClickDoc(show_subcommands=True)

        nav = self.add_nav("Development")

        page = nav.add_page("Changelog")
        page += mk.MkChangelog()

        page = nav.add_page("Code of conduct")
        page += mk.MkCodeOfConduct()

        page = nav.add_page("Contributing")
        page += mk.MkCommitConventions()
        page += mk.MkPullRequestGuidelines()

        page = nav.add_page("Setting up the environment")
        page += mk.MkDevEnvSetup()
        page += mk.MkDevTools(header="Tools")

        page = nav.add_page("Dependencies")
        page += mk.MkDependencyTable()
        page += mk.MkPipDepTree(direction="LR")

        page = nav.add_page("Module overview")
        page += mk.MkModuleOverview()

        if "mkdocs.plugins" in self.ctx.metadata.entry_points:
            page = nav.add_page("MkDocs Plugins")
            page += mk.MkPluginFlow()

        node = mk.MkLicense()
        page = nav.add_page("License", hide="toc")
        page += node

    @classmethod
    def for_project(cls, project, **kwargs):
        root = cls(context=project.context, **kwargs)
        project.set_root(root)
        return root


if __name__ == "__main__":
    theme = mk.MaterialTheme()
    doc = MkDefaultWebsite.for_project(
        mk.Project(theme=theme),
        static_pages={
            "Usage": "https://raw.githubusercontent.com/mkdocs/mkdocs/master/docs/getting-started.md",
        },
    )
    print(doc)



from __future__ import annotations

import os
import types

from typing import Any

from mknodes.info import grifferegistry
from mknodes.pages import mktemplatepage
from mknodes.utils import classhelpers, log, reprhelpers


DEFAULT_TPL = "modulepage.md"

logger = log.get_logger(__name__)


class MkModulePage(mktemplatepage.MkTemplatePage):
    """Page showing information about a module."""

    def __init__(
        self,
        module: tuple[str, ...] | str | types.ModuleType,
        *,
        klasses: list[type] | set[type] | None = None,
        path: str | os.PathLike | None = None,
        template: str | os.PathLike | None = None,
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            module: ModuleType or path to model to show info for.
            klasses: klasses to use
            path: Filename/path for the Module page. defaults to [modulename].md
            template: Name of the template to load
            kwargs: further keyword arguments passed to parent
        """
        self.parts = classhelpers.to_module_parts(module)
        self.module = classhelpers.to_module(module)
        self.klasses = klasses or list(
            classhelpers.iter_classes(module=self.parts, module_filter=self.parts[0]),
        )
        tpl_name = template or DEFAULT_TPL
        super().__init__(
            template=tpl_name,
            template_parent=DEFAULT_TPL if tpl_name != DEFAULT_TPL else None,
            path=path or f"{self.parts[-1]}.md",
            **kwargs,
        )

    def __repr__(self):
        return reprhelpers.get_repr(self, module=self.module, path=self.path)

    @property
    def extra_variables(self) -> dict[str, Any]:
        variables: dict[str, Any] = dict(module=self.module, klasses=self.klasses)
        variables["griffe_obj"] = grifferegistry.get_module(self.module)
        return variables


if __name__ == "__main__":
    doc = MkModulePage(mktemplatepage)
    print(doc)



from __future__ import annotations

from abc import ABCMeta
from collections.abc import MutableMapping

from mknodes.info import packageinfo
from mknodes.utils import log, packagehelpers


logger = log.get_logger(__name__)


def get_info(mod_name: str) -> packageinfo.PackageInfo:
    """Return info for given module from registry.

    Arguments:
        mod_name: Name of the module
    """
    return registry.get_info(mod_name)


class PackageRegistry(MutableMapping, metaclass=ABCMeta):
    """Registry for PackageInfos.

    Used for caching all loaded Package information.
    """

    def __init__(self):
        self._packages: dict[str, packageinfo.PackageInfo] = {}

    def __getitem__(self, value):
        return self._packages.__getitem__(value)

    def __setitem__(self, index, value):
        self._packages[index] = value

    def __delitem__(self, index):
        del self._packages[index]

    def __repr__(self):
        return f"{type(self).__name__}()"

    def __iter__(self):
        return iter(self._packages.keys())

    def __len__(self):
        return len(self._packages)

    def get_info(self, mod_name: str) -> packageinfo.PackageInfo:
        """Get package information for given module.

        Arguments:
            mod_name: Name of the module
        """
        mapping = packagehelpers.get_package_map()
        pkg_name = mapping[mod_name][0] if mod_name in mapping else mod_name
        pkg_name = pkg_name.lower()
        if pkg_name not in self._packages:
            self._packages[pkg_name] = packageinfo.PackageInfo(pkg_name)
        return self._packages[pkg_name]

    @property
    def inventory_urls(self) -> set[str]:
        """Return a set of inventory urls for all loaded packages."""
        return {v.inventory_url for v in self.values() if v.inventory_url is not None}


registry = PackageRegistry()


if __name__ == "__main__":
    reg = PackageRegistry()
    info = reg.get_info("mknodes")



from __future__ import annotations

from mknodes.basenodes import mkbasetable
from mknodes.utils import log, resources


logger = log.get_logger(__name__)


class MkTable(mkbasetable.MkBaseTable):
    """Class representing a formatted table."""

    REQUIRED_EXTENSIONS = [resources.Extension("tables")]

    def _to_markdown(self) -> str:
        table_data = self.data  # property
        if not any(table_data[k] for k in table_data):
            return ""
        widths = [self.width_for_column(c) for c in table_data]
        formatters = [f"{{:<{width}}}" for width in widths]
        headers = [formatters[i].format(k) for i, k in enumerate(table_data.keys())]
        divider = [width * "-" for width in widths]
        data = [
            [
                formatters[i].format(str(k).replace("\n", "<br>"))
                for i, k in enumerate(row)
            ]
            for row in self.iter_rows()
        ]
        header_txt = "| " + " | ".join(headers) + " |"
        divider_text = "| " + " | ".join(divider) + " |"
        data_txt = ["| " + " | ".join(line) + " |" for line in data]
        return "\n".join([header_txt, divider_text, *data_txt]) + "\n"

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        node_1 = MkTable(data={"Column A": ["A", "B", "C"], "Column B": ["C", "D", "E"]})
        # data can be given in different shapes.
        page += mk.MkReprRawRendered(node_1, header="### From dict")
        dicts = [{"col 1": "abc", "col 2": "cde"}, {"col 1": "fgh", "col 2": "ijk"}]
        node_2 = MkTable(data=dicts)
        page += mk.MkReprRawRendered(node_2, header="### From list")


if __name__ == "__main__":
    table = MkTable(data={"Column A": ["A", "B", "C"], "Column B": ["C", "D", "E"]})
    print(table)



from __future__ import annotations

from mknodes.info import packageregistry


def test_packageinfo():
    info = packageregistry.get_info("mknodes")
    assert info.license_name == "MIT"
    assert info.keywords
    assert info.extras is not None
    for package in info.required_package_names:
        packageregistry.get_info(package)



from __future__ import annotations

from collections.abc import Iterable
import functools
import re

from typing import TYPE_CHECKING, Any

from mknodes.basenodes import processors
from mknodes.data import datatypes
from mknodes.info import contexts
from mknodes.jinja import nodeenvironment
from mknodes.nodemods.modmanager import ModManager
from mknodes.treelib import node
from mknodes.utils import icons, log, resources


if TYPE_CHECKING:
    import mknodes as mk


logger = log.get_logger(__name__)


HEADER_REGEX = re.compile(r"^(#{1,6}) (.*)")


class MkNode(node.Node):
    """Base class for everything which can be expressed as Markup.

    The class inherits from Node. The idea is that starting from the
    root nav (aka Docs) down to nested Markup blocks, the whole project can be represented
    by one tree.

    MkNode is the base class for all nodes. We dont instanciate it directly.
    All subclasses carry an MkAnnotations node (except the MkAnnotations node itself)
    They can also pass an `indent` as well as a `shift_header_levels` keyword argument
    in order to modify the resulting markdown.
    """

    # METADATA (should be set by subclasses)

    ICON = "material/puzzle-outline"
    ATTR_LIST_SEPARATOR = " "
    REQUIRED_EXTENSIONS: list[resources.Extension] = []
    REQUIRED_PLUGINS: list[resources.Plugin] = []
    STATUS: datatypes.PageStatusStr | None = None
    CSS: list[resources.CSSFile | resources.CSSText] = []
    JS_FILES: list[resources.JSFile] = []
    children: list[MkNode]
    _name_registry: dict[str, MkNode] = dict()

    def __init__(
        self,
        *,
        header: str = "",
        indent: str = "",
        name: str | None = None,
        shift_header_levels: int = 0,
        css_classes: Iterable[str] | None = None,
        as_html: bool = False,
        context: contexts.ProjectContext | None = None,
        parent: MkNode | None = None,
    ):
        """Constructor.

        Arguments:
            header: Optional header for contained Markup
            indent: Indentation of given Markup (unused ATM)
            name: An optional unique identifier (allows getting node via MkNode.get_node)
            shift_header_levels: Regex-based header level shifting (adds/removes #-chars)
            css_classes: A sequence of css class names to use for this node
            as_html: Converts node to HTML on stringifying.
            context: Context this Nav is connected to.
            parent: Parent for building the tree
        """
        super().__init__(parent=parent)
        self.header = header
        self.indent = indent
        self.shift_header_levels = shift_header_levels
        self._files: dict[str, str | bytes] = {}
        self.mods = ModManager()
        self.mods._css_classes = list(css_classes) if css_classes else []
        self._ctx = context
        self._node_name = name
        self.as_html = as_html
        if name is not None:
            self._name_registry[name] = self
        # ugly, but convenient.
        import mknodes as mk

        if not isinstance(self, mk.MkAnnotations):
            self.annotations = mk.MkAnnotations(parent=self)
        else:
            self.annotations = None

    @functools.cached_property
    def env(self):
        """The node jinja environment.

        The environment has additional loaders for the class file path
        as well as the resolved parent nav file path.
        """
        return nodeenvironment.NodeEnvironment(self)

    def __str__(self):
        return self.to_html() if self.as_html else self.to_markdown()

    def __hash__(self):
        return hash(self.to_markdown())

    def __eq__(self, other):
        if not type(other) == type(self):
            return False
        dct_1 = self.__dict__.copy()
        dct_2 = other.__dict__.copy()
        for attr in ["_parent"]:  # , "_annotations"]
            dct_1.pop(attr)
            dct_2.pop(attr)
        return dct_1 == dct_2

    def __rshift__(self, other, inverse: bool = False):
        import mknodes as mk

        if self.parent or (isinstance(other, mk.MkNode) and other.parent):
            msg = "Can only perform shift when nodes have no parent"
            raise RuntimeError(msg)
        container = mk.MkContainer(parent=self.parent, block_separator=" ")
        if inverse:
            container.append(other)
            container.append(self)
        else:
            container.append(self)
            container.append(other)
        return container

    def __rrshift__(self, other):
        return self.__rshift__(other, inverse=True)

    # @property
    # def ctx(self):
    #     """The tree context."""
    #     if root := self.ctx_root:
    #         return root._ctx
    #     return contexts.ProjectContext()

    @property
    def ctx(self):
        """The tree context."""
        if self._ctx:
            return self._ctx
        for ancestor in self.ancestors:
            if ancestor._ctx:
                return ancestor._ctx
        return contexts.ProjectContext()

    @property
    def parent_navs(self) -> list[mk.MkNav]:
        """Return a list of parent MkNavs, ordered from root to leaf."""
        import mknodes as mk

        navs = [i for i in self.ancestors if isinstance(i, mk.MkNav)]
        return list(reversed(navs))

    @property
    def parent_page(self) -> mk.MkPage | None:
        import mknodes as mk

        return next((i for i in self.ancestors if isinstance(i, mk.MkPage)), None)

    def to_child_node(self, other: Any):  # type: ignore[return]
        """Convert given nodes / strings to child nodes.

        Either converts text to an MkNode sets parent of node to self.

        Arguments:
            other: The node / string to convert to a child node.
        """
        import mknodes as mk

        match other:
            case str() if (match := HEADER_REGEX.match(other)) and "\n" not in other:
                return mk.MkHeader(match[2], level=len(match[1]), parent=self)
            case str():
                return mk.MkText(other, parent=self)
            case list():
                return mk.MkContainer([self.to_child_node(i) for i in other])
            case mk.MkNode():
                other.parent = self
                return other
            case _:
                raise TypeError(other)

    @classmethod
    def get_node(cls, name: str) -> MkNode:
        """Get a node from name registry."""
        return cls._name_registry[name]

    def _to_markdown(self) -> str:
        return NotImplemented

    def to_markdown(self) -> str:
        """Outputs markdown for self and all children."""
        text = self._to_markdown()
        for proc in self.get_processors():
            text = proc.run(text)
        return text

    def get_processors(self) -> list[processors.TextProcessor]:
        """Return list of processors used to created markdown."""
        return [
            processors.ShiftHeaderLevelProcessor(self.shift_header_levels),
            processors.IndentationProcessor(self.indent),
            processors.AppendCssClassesProcessor(self),
            processors.PrependHeaderProcessor(self.header),
            processors.AnnotationProcessor(self),
        ]

    def attach_annotations(self, text: str) -> str:
        """Attach annotations block to given markdown.

        Can be reimplemented if non-default annotations are needed.

        Arguments:
            text: Markdown to annote
        """
        return self.annotations.annotate_text(text) if self.annotations else text

    def attach_css_classes(self, text: str) -> str:
        """Attach CSS classes to given markdown in attr-list style.

        Can be reimplemented if non-default behavior is needed.
        Default behavior is appending the css class snippet with a space
        as separator.

        Arguments:
            text: Markdown to annote
        """
        if not self.mods.css_classes:
            return text
        classes = " ".join(f".{kls_name}" for kls_name in self.mods.css_classes)
        text += f"{self.ATTR_LIST_SEPARATOR}{{: {classes}}}"
        return text

    @property
    def resolved_parts(self) -> tuple[str, ...]:
        """Return a tuple containing all section names."""
        import mknodes as mk

        parts = [nav.section for nav in self.parent_navs if nav.section]
        if isinstance(self, mk.MkNav) and self.section:
            parts.append(self.section)
        return tuple(parts)

    @property
    def files(self):
        """Return a dict containing the virtual files attached to this tree element.

        This can be overridden by nodes if they want files to be included dynamically.
        For static files, use `add_file`.
        """
        return self._files

    def add_file(self, filename: str, data: str | bytes):
        """Add a static file as data to this node.

        Arguments:
            filename: Filename of the file to add
            data: Data of the file
        """
        self._files[filename] = data

    def add_css_class(self, class_name: str):
        """Wrap node markdown with given css class.

        Arguments:
            class_name: CSS class to wrap the node with
        """
        self.mods._css_classes.append(class_name)

    def get_node_resources(self) -> resources.Resources:
        """Return the resources specific for this node."""
        extension = {k.extension_name: dict(k) for k in self.REQUIRED_EXTENSIONS}
        mod_resources = self.mods.get_resources()
        css_resources: list[resources.CSSType] = []
        env = self.env
        for css in self.CSS + mod_resources.css:
            if isinstance(css, resources.CSSFile) and css.is_local():
                text = env.render_template(css.link)
                css_resource = resources.CSSText(text, css.link)
                css_resources.append(css_resource)
            else:
                css_resources.append(css)
        js_resources: list[resources.JSType] = []
        for js_file in self.JS_FILES + mod_resources.js:
            if isinstance(js_file, resources.JSFile) and js_file.is_local():
                text = env.render_template(js_file.link)
                js_resource = resources.JSText(
                    text,
                    js_file.link,
                    defer=js_file.defer,
                    async_=js_file.async_,
                    crossorigin=js_file.crossorigin,
                    typ=js_file.typ,
                    is_library=js_file.is_library,
                )
                js_resources.append(js_resource)
            else:
                js_resources.append(js_file)
        return resources.Resources(
            js=js_resources,
            markdown_extensions=extension,
            plugins=self.REQUIRED_PLUGINS,
            css=css_resources,
        )

    def get_resources(self) -> resources.Resources:
        """Return the "final" resources object."""
        nodes = [*list(self.descendants), self]
        extensions: dict[str, dict] = {
            "attr_list": {},
            "md_in_html": {},
            "pymdownx.emoji": {
                "emoji_index": icons.twemoji,
                "emoji_generator": icons.to_svg,
            },
            "pymdownx.magiclink": dict(
                repo_url_shorthand=True,
                user=self.ctx.metadata.repository_username,
                repo=self.ctx.metadata.repository_name,
            ),
        }

        req = resources.Resources(markdown_extensions=extensions)
        for _node in nodes:
            node_req = _node.get_node_resources()
            req.merge(node_req)
        return req

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        # We dont instanciate MkNode directly, so we take a subclass
        # to show some base class functionality

        node = mk.MkText("Intro\n# A header\nOutro")
        node.shift_header_levels = 2
        page += mk.MkReprRawRendered(node, header="### Shift header levels")

        node = mk.MkText("Every node can also append annotations (1)")
        node.annotations[1] = "Nice!"
        page += mk.MkReprRawRendered(node, header="### Append annotations")

    @classmethod
    def with_default_context(cls, *args, **kwargs):
        import mknodes as mk

        theme = mk.MaterialTheme()
        proj = mk.Project(theme=theme)
        return cls(*args, **kwargs, context=proj.context)

    def to_html(self) -> str:
        """Convert node to HTML using the resources from node + children."""
        import markdown

        md = self.to_markdown()
        reqs = self.get_resources()
        configs = reqs.markdown_extensions
        exts = list(configs.keys())
        return markdown.Markdown(extensions=exts, extension_configs=configs).convert(md)


if __name__ == "__main__":
    import mknodes as mk

    section = "pre" >> mk.MkText("hello\n# Header\nfdsfds") >> "test" >> "xx"
    print(section)



from __future__ import annotations

from typing import Any, Literal

from mknodes.basenodes import mknode
from mknodes.utils import icons, log, reprhelpers, resources, xmlhelpers as xml


logger = log.get_logger(__name__)


class MkMaterialBadge(mknode.MkNode):
    """Node for a CSS-based badge a la MkDocs-Material."""

    ICON = "simple/shieldsdotio"
    CSS = [resources.CSSFile("css/materialbadge.css")]

    def __init__(
        self,
        icon: str,
        text: str = "",
        typ: Literal["heart", "right", ""] = "",
        link: str | None = None,
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            icon: Icon to display
            text: Text to display
            typ: Optional badge type
            link: Optional URL to link to
            kwargs: Keyword arguments passed to parent
        """
        super().__init__(**kwargs)
        self.icon = icon
        self.text = text
        self.typ = typ
        self.link = link

    def _to_markdown(self):
        classes = (
            f"md-typeset mdx-badge mdx-badge--{self.typ}"
            if self.typ
            else "mdx-badge md-typeset"
        )
        root = xml.Span(classes)
        if self.icon:
            icon = icons.get_emoji_slug(self.icon)
            xml.Span("mdx-badge__icon", parent=root, text=icon)
        if self.text:
            xml.Span("mdx-badge__text", parent=root, text=self.text)
        return root.to_string()

    def __repr__(self):
        return reprhelpers.get_repr(
            self,
            icon=self.icon,
            text=self.text,
            typ=self.typ,
            link=self.link,
            _filter_empty=True,
        )

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        node = MkMaterialBadge(icon=":material-file:", text="text")
        page += mk.MkReprRawRendered(node)

        node = MkMaterialBadge(icon=":material-file:", text="text", typ="heart")
        page += mk.MkReprRawRendered(node)


if __name__ == "__main__":
    img = MkMaterialBadge("mdi:wrench", "right")
    print(img)



from __future__ import annotations

import mknodes as mk


EXPECTED_TABBED = """===! "Tab1"
    Some text

=== "Tab2"
    Another text
"""


def test_mktabbed():
    tabs = dict(Tab1="Some text", Tab2="Another text")
    node = mk.MkTabbed(tabs)
    assert len(node.items) == 2  # noqa: PLR2004
    # assert str(node["Tab1"]) == "Some text"
    # assert str(node[1]) == "Another text"
    assert str(node) == EXPECTED_TABBED
    node["Tab3"] = "And another one"



from __future__ import annotations

from collections.abc import Mapping
from typing import Any

from mknodes.basenodes import mkcontainer, mknode
from mknodes.utils import log, reprhelpers, resources


logger = log.get_logger(__name__)


class MkDefinition(mkcontainer.MkContainer):
    """Node for a single definition."""

    REQUIRED_EXTENSIONS = [resources.Extension("def_list")]
    ICON = "material/library"

    def __init__(
        self,
        title: str = "",
        content: list | None | str | mknode.MkNode = None,
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            title: Setting title
            content: Markdown content for this block
            kwargs: Keyword arguments passed to parent
        """
        super().__init__(content=content, **kwargs)
        self._title = title

    @property
    def title(self):
        return self._title

    @title.setter
    def title(self, value):
        self._title = value

    def __repr__(self):
        return reprhelpers.get_repr(self, title=self._title, content=self.items)

    def _to_markdown(self) -> str:
        lines = super()._to_markdown().split("\n")
        result = [f"{self.title}", f":   {lines[0]}"]
        result.extend(f"    {i}" for i in lines[1:])
        result.append("")
        return "\n".join(result) + "\n"

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        node = MkDefinition("test", content="hfkdlsjk")
        page += mk.MkReprRawRendered(node, header="### Regular")


class MkDefinitionList(mkcontainer.MkContainer):
    """Node for definition lists."""

    REQUIRED_EXTENSIONS = [resources.Extension("def_list")]
    ICON = "material/library"

    def __init__(
        self,
        data: Mapping | None = None,
        *,
        header: str = "",
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            data: Data show for the table
            header: Section header
            kwargs: Keyword arguments passed to parent
        """
        super().__init__(header=header, **kwargs)
        self.data: dict[str, str | mknode.MkNode] = {}
        self.items = data

    def __repr__(self):
        kwarg_data = {k: reprhelpers.to_str_if_textnode(v) for k, v in self.data.items()}
        return reprhelpers.get_repr(self, data=kwarg_data)

    @property
    def items(self):
        return list(self.data.values())

    @items.setter
    def items(self, data):
        match data:
            case Mapping():
                self.data = {k: self.to_child_node(v) for k, v in data.items()}
            case list():
                self.data = {
                    str(i): self.to_child_node(item) for i, item in enumerate(data)
                }
            case None:
                self.data = {}
            case _:
                raise TypeError(data)

    def _to_markdown(self) -> str:
        result = []
        for k, v in self.data.items():
            lines = str(v).split("\n")
            result.extend([str(k), f":   {lines[0]}"])
            result.extend(f"    {i}" for i in lines[1:])
            result.append("")
        return "\n".join(result) + "\n"

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        defs_1 = dict(something="A definition", somethingelse="Another\ndefinition")
        node = MkDefinitionList(data=defs_1)
        page += mk.MkReprRawRendered(node, header="### Regular")

        defs_2 = dict(admonition=mk.MkAdmonition("Nested markup"))
        node = MkDefinitionList(data=defs_2)
        page += mk.MkReprRawRendered(node, header="### Nested markup")


if __name__ == "__main__":
    ls = MkDefinitionList(data=dict(a="b\nc", c="d"))
    print(ls)



from __future__ import annotations

import contextlib
import functools
import io
import os

from typing import Any, Literal

from git_changelog import cli

from mknodes.basenodes import mktext
from mknodes.utils import log, reprhelpers, resources


logger = log.get_logger(__name__)


@functools.cache
def get_changelog(
    repository: str,
    template: str,
    convention: str,
    sections: tuple[str, ...] | None = None,
) -> str:
    with contextlib.redirect_stdout(io.StringIO()):
        _changelog, text = cli.build_and_render(
            repository=repository,
            template=template,
            convention=convention,
            parse_refs=True,
            parse_trailers=True,
            sections=list(sections) if sections else None,
        )
    return text


class MkChangelog(mktext.MkText):
    """Node for a git-based changelog (created by git-changelog).

    !!! note
        For building a changelog with Github Actions, the actions/checkout@v3
        action needs to have fetch-depth set to 0 (or some other value.)
    """

    ICON = "material/format-list-group"
    REQUIRED_PACKAGES = [resources.Package("git-changelog")]

    def __init__(
        self,
        convention: Literal["basic", "angular", "atom", "conventional"] = "conventional",
        template: Literal["keepachangelog", "angular"] = "keepachangelog",
        sections: list[str] | None = None,
        repository: str | os.PathLike | None = None,
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            convention: Commit conventions to use
            template: Changelog template
            sections: Which sections to display
            repository: git repo to use for changelog (defaults to current folder)
            kwargs: Keyword arguments passed to parent
        """
        super().__init__(**kwargs)
        self.convention = convention
        self.template = template
        self.sections = sections
        self._repository = repository

    def __repr__(self):
        return reprhelpers.get_repr(
            self,
            convention=self.convention,
            template=self.template,
            sections=self.sections,
            repository=str(self.repository),
            _filter_empty=True,
        )

    @property
    def repository(self):
        match self._repository:
            case None:
                return self.ctx.metadata.repository_path
            case _:
                return self._repository

    @property
    def text(self) -> str:
        return get_changelog(
            repository=str(self.repository),
            template=self.template,
            convention=self.convention,
            sections=tuple(self.sections) if self.sections else None,
        )

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        node = MkChangelog(template="keepachangelog", shift_header_levels=2)
        page += mk.MkReprRawRendered(node, header="### keepachangelog template")
        node = MkChangelog(template="angular", shift_header_levels=2)
        page += mk.MkReprRawRendered(node, header="### angular template")
        node = MkChangelog(convention="basic", shift_header_levels=2)
        page += mk.MkReprRawRendered(node, header="### basic convention")


if __name__ == "__main__":
    changelog = MkChangelog()
    print(changelog)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal, get_args

from mknodes.basenodes import mkcontainer
from mknodes.utils import log, reprhelpers, resources


if TYPE_CHECKING:
    import mknodes as mk

logger = log.get_logger(__name__)

CriticMarkStr = Literal["addition", "deletion", "comment", "highlight"]


class MkCritic(mkcontainer.MkContainer):
    """MkCritic block."""

    ICON = "material/format-text"
    REQUIRED_EXTENSIONS = [resources.Extension("pymdownx.critic")]

    def __init__(
        self,
        content: str | mk.MkNode | list,
        *,
        mark: CriticMarkStr = "highlight",
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            content: Content to mark
            mark: type of mark
            kwargs: Keyword arguments passed to parent
        """
        super().__init__(content=content, **kwargs)
        self.mark = mark

    def __repr__(self):
        return reprhelpers.get_repr(self, content=self.items, mark=self.mark)

    def _to_markdown(self) -> str:
        match self.mark:
            case "addition":
                left, right = ("++", "++")
            case "deletion":
                left, right = ("--", "--")
            case "highlight":
                left, right = ("==", "==")
            case "comment":
                left, right = (">>", "<<")
            case _:
                raise TypeError(self.mark)
        return f"{{{left}\n\n{super()._to_markdown()}\n\n{right}}}"

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        page += "The MkCritic node can be used to display text diffs."
        for typ in get_args(CriticMarkStr):
            node = MkCritic(f"This is type {typ}", mark=typ)
            page += mk.MkHeader(f"Type {typ!r}", level=3)
            page += mk.MkReprRawRendered(node)


if __name__ == "__main__":
    mkcritic = MkCritic("hello")
    print(mkcritic)



from __future__ import annotations

from collections.abc import Mapping


class CSS:
    def __init__(self, obj: Mapping):
        self.obj = obj
        self._data: dict[str, list] = {}
        self._parse(obj)

    def __repr__(self):
        return self.to_string()

    def to_string(self) -> str:
        string = ""
        for key, value in sorted(self._data.items()):
            if self._data[key]:
                string += key[1:] + " {\n" + "".join(value) + "}\n\n"
        return string

    def _parse(self, obj, selector: str = ""):
        for key, value in obj.items():
            if hasattr(value, "items"):
                rule = selector + " " + key
                self._data[rule] = []
                self._parse(value, rule)

            else:
                prop = self._data[selector]
                prop.append(f"\t{key}: {value};\n")


if __name__ == "__main__":
    css = CSS({".test": {"color": "red"}})
    print(css)



from __future__ import annotations

from typing import Literal

from mknodes.basenodes import mktable
from mknodes.utils import layouts, log, reprhelpers


logger = log.get_logger(__name__)


class MkClassTable(mktable.MkTable):
    """Node for a table showing info for a list of classes."""

    def __init__(
        self,
        klasses: list[type] | set[type],
        *,
        layout: Literal["compact", "extended"] | layouts.Layout = "extended",
        **kwargs,
    ):
        self.klasses = klasses
        # STRIP_CODE = r"```[^\S\r\n]*[a-z]*\n.*?\n```"
        # docs = [re.sub(STRIP_CODE, '', k.__module__, 0, re.DOTALL) for k in klasses]
        self.layout = layout
        super().__init__(**kwargs)

    def __repr__(self):
        return reprhelpers.get_repr(self, layout=self.layout, klasses=self.klasses)

    @property
    def layouter(self):
        match self.layout:
            case "compact":
                return layouts.CompactClassLayout(link_provider=self.ctx.links)
            case "extended":
                return layouts.ExtendedClassLayout(link_provider=self.ctx.links)
            case layouts.Layout():
                return self.layout
            case _:
                raise ValueError(self.layout)

    @property
    def data(self):
        if not self.klasses:
            return {}
        layouter = self.layouter
        data = [layouter.get_row_for(kls) for kls in self.klasses]
        return {
            k: [self.to_child_node(dic[k]) for dic in data]  # type: ignore[index]
            for k in data[0]
        }

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        klasses = [mk.MkTable, MkClassTable, mk.MkNav]
        node_1 = MkClassTable(klasses=klasses, layout="compact")
        node_2 = MkClassTable(klasses=klasses, layout="extended")
        page += mk.MkReprRawRendered(node_1, header="### Compact layout")
        page += mk.MkReprRawRendered(node_2, header="### Extended layout")


if __name__ == "__main__":
    table = MkClassTable(klasses=[mktable.MkTable], layout="extended")
    print(table)



from __future__ import annotations

from collections.abc import Sequence
from importlib import metadata
import os
import sys
import types

from typing import TYPE_CHECKING

from mknodes import paths
from mknodes.info import packageregistry
from mknodes.utils import helpers, inventorymanager, log


logger = log.get_logger(__name__)


if TYPE_CHECKING:
    import mknodes as mk

    LinkableType = str | mk.MkPage | mk.MkNav | mk.MkHeader | types.ModuleType | type
    """A type which can get linked by the LinkProvider."""


def homepage_for_distro(dist_name: str) -> str | None:
    """Return a link for given distribution from metadata.

    Used if no better link was found.

    Arguments:
        dist_name: Name of the distribution to get a link for
    """
    if dist_name in sys.stdlib_module_names:
        return None
    try:
        dist = packageregistry.get_info(dist_name)
    except metadata.PackageNotFoundError:
        logger.debug("Could not get package info for %s", dist_name)
        return None
    else:
        return dist.homepage


def linked(identifier: str, title: str | None = None) -> str:
    """Create a markdown link.

    Arguments:
        identifier: Target url
        title: Title to show as label
    """
    suffix = "" if helpers.is_url(identifier) or identifier.endswith(".md") else ".md"
    return f"[{identifier if title is None else title}]({identifier}{suffix})"


class LinkProvider:
    def __init__(
        self,
        base_url: str = "",
        use_directory_urls: bool = True,
        include_stdlib: bool = False,
    ):
        """Constructor.

        Arguments:
            base_url: Base URL of the website
            use_directory_urls: Use directory-style URLS
            include_stdlib: Load the stdlib inventory file on init
        """
        self.inv_manager = inventorymanager.InventoryManager()
        self.base_url = base_url
        self.excludes: set[str] = set()
        self.use_directory_urls = use_directory_urls
        if include_stdlib:
            self.add_inv_file(
                paths.RESOURCES / "python_objects.inv",
                base_url="https://docs.python.org/3/",
            )

    def set_excludes(self, excludes: Sequence[str]):
        """Set terms which should not get picked up by the linkprovider.

        That way handling of linking can be done by Markdown extensions.

        Arguments:
            excludes: list of terms to exclude from linking
        """
        self.excludes.update(excludes)

    def add_inv_file(self, path: str | os.PathLike, base_url: str | None = None):
        """Add an inventory file to the inventory manager.

        Arguments:
            path: Path or URL to the inventory file
            base_url: Base URL (required when inventory file is local)
        """
        self.inv_manager.add_inv_file(path, base_url=base_url)

    def url_for_module(
        self,
        mod: types.ModuleType,
        fallback_to_homepage: bool = True,
    ) -> str | None:
        """Return a url for given module.

        Arguments:
            mod: Module to get a url for
            fallback_to_homepage: Whether to get a link from Metadata if no other found
        """
        dotted_path = mod.__name__
        if dotted_path in self.inv_manager:
            return self.inv_manager[dotted_path]
        module = dotted_path.split(".")[0]
        return homepage_for_distro(module) if fallback_to_homepage else None

    def link_for_module(self, mod: types.ModuleType) -> str:
        """Return a markdown link for given module.

        Arguments:
            mod: Module to get a link for
        """
        dotted_path = mod.__name__
        fallback = dotted_path not in self.excludes
        if url := self.url_for_module(mod, fallback_to_homepage=fallback):
            return linked(url, dotted_path)
        return linked(dotted_path)

    def url_for_klass(
        self,
        kls: type,
        fallback_to_homepage: bool = True,
    ) -> str | None:
        """Return a url for given class.

        Arguments:
            kls: Klass to get a url for
            fallback_to_homepage: Whether to get a link from Metadata if no other found
        """
        module_path = kls.__module__
        qual_name = kls.__qualname__.split("[")[0]  # to split off generics part
        is_builtin = module_path == "builtins"
        dotted_path = qual_name if is_builtin else f"{module_path}.{qual_name}"
        if dotted_path in self.inv_manager:
            return self.inv_manager[dotted_path]
        module = module_path.split(".")[0]
        return homepage_for_distro(module) if fallback_to_homepage else None

    def link_for_klass(self, kls: type) -> str:
        """Return a markdown link for given class.

        Arguments:
            kls: Klass to get a link for
        """
        qual_name = kls.__qualname__.split("[")[0]  # to split off generics part
        fallback = qual_name not in self.excludes
        if url := self.url_for_klass(kls, fallback_to_homepage=fallback):
            return linked(url, qual_name)
        return linked(qual_name)

    def url_for_nav(self, nav: mk.MkNav) -> str:
        """Return the final URL for given MkNav.

        Arguments:
            nav: The Nav to link to
        """
        if nav.index_page:
            path = nav.index_page.resolved_file_path
            path = path.replace(".md", ".html")
        else:
            path = nav.resolved_file_path
        return self.base_url + path

    def url_for_page(self, page: mk.MkPage) -> str:
        """Return the final URL for given MkPage.

        Arguments:
            page: The Page to link to
        """
        path = page.resolved_file_path
        if self.use_directory_urls:
            path = path.replace(".md", "/")
        else:
            path = path.replace(".md", ".html")
        return self.base_url + path

    def url_for_header(self, header: mk.MkHeader) -> str:
        """Return the final URL for given MkHeader.

        Arguments:
            header: The Header to link to
        """
        page = header.parent_page
        if page is None:
            msg = "Need a parent page for MkHeader in order to link to it"
            raise RuntimeError(msg)
        suffix = "#" + helpers.slugify(header.text)
        return self.url_for_page(page) + suffix

    def get_link(self, target: LinkableType, title: str | None = None):
        """Return a markdown link for given target.

        Target can be a class, a module, a MkPage, MkNav or a string.

        Arguments:
            target: The thing to link to
            title: The title to use for the link
        """
        url = self.get_url(target)
        return linked(url, title)

    def get_url(self, target: LinkableType) -> str:  # type: ignore  # noqa: PLR0911
        """Get a url for given target.

        Target can be a class, a module, a MkPage, MkNav or a string.

        Arguments:
            target: The thing to link to
        """
        import mknodes as mk

        match target:
            case mk.MkPage():
                return self.url_for_page(target)
            case mk.MkNav():
                return self.url_for_nav(target)
            case mk.MkHeader():
                return self.url_for_header(target)
            case type():
                return self.url_for_klass(target) or ""
            case types.ModuleType():
                return self.url_for_module(target) or ""
            case str() if target.startswith("/"):
                return self.base_url.rstrip("/") + target
            case str() if helpers.is_url(target):
                return target
            case str() if target in mk.MkNode._name_registry:
                node = mk.MkNode.get_node(target)
                if isinstance(node, mk.MkPage | mk.MkNav):
                    return self.get_url(node)
                msg = f"Cannot get link for {node!r}"
                raise TypeError(msg)
            case str():
                return f"{target}.md"
            case _:
                raise TypeError(target)


if __name__ == "__main__":
    provider = LinkProvider()
    provider.add_inv_file("https://docs.python.org/3/objects.inv")
    link = provider.link_for_klass(LinkProvider)
    print(link)



from __future__ import annotations

import argparse

from mknodes.basenodes import mkcontainer, mkheader, mktext
from mknodes.utils import log, reprhelpers


logger = log.get_logger(__name__)


class MkArgParseHelp(mkcontainer.MkContainer):
    """Node to describe an argparse parser."""

    ICON = "material/bash"
    STATUS = "new"

    def __init__(self, parser, **kwargs):
        self.parser = parser
        super().__init__(**kwargs)

    def __repr__(self):
        return reprhelpers.get_repr(self, parser=self.parser)

    @property
    def items(self):
        # ArgumentParser.format_usage()
        # ArgumentParser.format_help()
        items = [
            mkheader.MkHeader(self.parser.prog),
            mktext.MkText(self.parser.description),
            mktext.MkText("Options:"),
        ]
        lines = []
        for action in self.parser._actions:
            opts = [f"`{opt}`" for opt in action.option_strings]
            if not opts:
                continue
            line = "- " + ",".join(opts)
            if action.metavar:
                line += f" `{action.metavar}`"
            line += f": {action.help}".replace(" Default: %(default)s", "")
            if action.default and action.default != argparse.SUPPRESS:
                line += f"(default: `{action.default}`)"
            lines.append(line)
        items.append(mktext.MkText("\n".join(lines)))
        for i in items:
            i.parent = self
        return items

    @items.setter
    def items(self, value):
        pass

    @classmethod
    def create_example_page(cls, page):
        from git_changelog import cli

        import mknodes as mk

        parser = cli.get_parser()

        node = MkArgParseHelp(parser)
        page += mk.MkReprRawRendered(node)


if __name__ == "__main__":
    from git_changelog import cli

    parser = cli.get_parser()

    text = MkArgParseHelp(parser)
    print(text)



from __future__ import annotations

from typing import Any

from mknodes.basenodes import mkcontainer, mknode
from mknodes.data import buildsystems
from mknodes.utils import log, reprhelpers


logger = log.get_logger(__name__)

EXAMPLE_URL = "http://www.some-github-provider.com/my-project.git"

START_TEXT = """All development for this library happens in the
{link} repo on GitHub.
First, you'll need to download the source code and install an
editable version of the Python package:"""

CLONE_CODE = """
# Clone the repository
git clone {repo_url}
cd {folder_name}
"""


def get_build_backend_section(backend: buildsystems.BuildSystem) -> list[mknode.MkNode]:
    import mknodes as mk

    backend_name = backend.identifier.capitalize()
    return [
        mk.MkHeader("Build system"),
        mk.MkText(f"{backend_name} is used as the build system."),
        mk.MkCode(f"pip install {backend.identifier}", language="bash"),
        mk.MkLink(backend.url, "More information"),
    ]


class MkDevEnvSetup(mkcontainer.MkContainer):
    """Text node containing Instructions to set up a dev environment."""

    ICON = "material/dev-to"
    STATUS = "new"

    def __init__(
        self,
        *,
        repo_url: str | None = None,
        build_backend: buildsystems.BuildSystemStr | None = None,
        header: str = "Setting up a development environment",
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            repo_url: Repo url to show. If None, it will be pulled from project.
            build_backend: Build backend to show install instructions for.
                            If None, it will be pulled from project.
            header: Section header
            kwargs: Keyword arguments passed to parent
        """
        super().__init__(header=header, **kwargs)
        self._repo_url = repo_url
        self._build_backend = build_backend

    def __repr__(self):
        return reprhelpers.get_repr(
            self,
            repo_url=self._repo_url,
            build_backend=self._build_backend,
            _filter_empty=True,
        )

    @property
    def repo_url(self) -> str:
        match self._repo_url:
            case None:
                repo_url = self.ctx.metadata.repository_url
            case str():
                repo_url = self._repo_url
            case _:
                raise TypeError(self._repo_url)
        repo_url = repo_url.rstrip("/")
        if not repo_url.endswith(".git"):
            repo_url += ".git"
        return repo_url

    @repo_url.setter
    def repo_url(self, value):
        self._repo_url = value

    @property
    def build_backend(self) -> buildsystems.BuildSystem:  # type: ignore[return]
        if self._build_backend is None:
            return self.ctx.metadata.build_system or buildsystems.setuptools
        return buildsystems.BUILD_SYSTEMS[self._build_backend]

    @build_backend.setter
    def build_backend(self, value):
        self._build_backend = value

    @property
    def items(self):
        import mknodes as mk

        folder_name = self.repo_url.removesuffix(".git").split("/")[-1]
        code = CLONE_CODE.format(repo_url=self.repo_url, folder_name=folder_name)
        link = mk.MkLink(self.repo_url, folder_name)
        start_text = START_TEXT.format(link=str(link))
        items = [mk.MkText(start_text), mk.MkCode(code, language="md")]
        items.extend(get_build_backend_section(self.build_backend))
        for item in items:
            item.parent = self
        return items

    @items.setter
    def items(self, value):
        pass

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        node = MkDevEnvSetup(header="")
        page += mk.MkReprRawRendered(node, header="### From project")
        node = MkDevEnvSetup(header="", repo_url="http://url_to_git_repo.com/name.git")
        page += mk.MkReprRawRendered(node, header="### Explicit")


if __name__ == "__main__":
    setup_text = MkDevEnvSetup(build_backend="flit")
    print(setup_text)



from __future__ import annotations

from typing import Any, Literal

from mknodes.basenodes import mknode
from mknodes.utils import log, reprhelpers, resources


logger = log.get_logger(__name__)


class MkProgressBar(mknode.MkNode):
    """Node to display a CSS-based progress bar."""

    REQUIRED_EXTENSIONS = [resources.Extension("pymdownx.progressbar")]
    ICON = "fontawesome/solid/bars-progress"
    STATUS = "new"
    CSS = [resources.CSSFile("css/progressbar.css")]
    ATTR_LIST_SEPARATOR = ""

    def __init__(
        self,
        percentage: int,
        title: str | None | Literal[True] = True,
        style: Literal["thin", "candystripe", "candystripe_animated"] | None = None,
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            percentage: Percentage value for the progress bar
            title: Title to display on top of progress bar
            style: Progress bar style
            kwargs: Keyword arguments passed to parent
        """
        super().__init__(**kwargs)
        self.title = title
        self.percentage = percentage
        self.style = style
        match self.style:
            case "thin":
                self.add_css_class("thin")
            case "candystripe":
                self.add_css_class("candystripe")
            case "candystripe_animated":
                self.add_css_class("candystripe")
                self.add_css_class("candystripe-animate")

    def __repr__(self):
        return reprhelpers.get_repr(
            self,
            percentage=self.percentage,
            title=self.title,
            style=self.style,
        )

    def _to_markdown(self) -> str:
        match self.title:
            case str():
                title = self.title.format(percentage=self.percentage)
            case True:
                title = f"{self.percentage}%"
            case _:
                title = ""
        return rf'[={self.percentage}% "{title}"]'

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        node = MkProgressBar(60)
        page += mk.MkReprRawRendered(node, header="### Regular")
        node = MkProgressBar(60, style="thin")
        page += mk.MkReprRawRendered(node, header="### Thin")
        node = MkProgressBar(70, style="candystripe", title="We reached {percentage}!")
        page += mk.MkReprRawRendered(node, header="### Candystripe")
        node = MkProgressBar(80, style="candystripe_animated")
        page += mk.MkReprRawRendered(node, header="### Animated")


if __name__ == "__main__":
    bar = MkProgressBar(percentage=30, title=None)
    print(bar)



from __future__ import annotations

from collections.abc import Callable
from dataclasses import Field
import types

from typing import Any, ClassVar, Literal, Protocol


AdmonitionTypeStr = Literal[
    "note",
    "abstract",
    "info",
    "tip",
    "success",
    "question",
    "warning",
    "failure",
    "danger",
    "bug",
    "example",
    "quote",
]

ClassifierStr = Literal[
    "Development Status",
    "Environment",
    "Framework",
    "Intended Audience",
    "License",
    "Natural Language",
    "Operating System",
    "Programming Language",
    "Topic",
    "Typing",
]

CLASSIFIERS: list[ClassifierStr] = [
    "Development Status",
    "Environment",
    "Framework",
    "Intended Audience",
    "License",
    "Natural Language",
    "Operating System",
    "Programming Language",
    "Topic",
    "Typing",
]

EXT_TO_PYGMENTS_STYLE = {
    ".py": "py",
    ".yml": "yaml",
    ".yaml": "yaml",
    ".toml": "toml",
    ".md": "md",
    ".jinja": "jinja",
}

MetadataTypeStr = (
    Literal[
        "classifiers",
        "keywords",
        "keywords_combined",
        "websites",
        "dependencies",
        "required_python",
        "installed_packages",
    ]
    | ClassifierStr
)


class DataclassInstance(Protocol):
    __dataclass_fields__: ClassVar[dict[str, Field[Any]]]


PageStatusStr = Literal["new", "deprecated", "encrypted"]

RGBColorType = tuple[int, int, int] | str
RGBAColorType = tuple[int, int, int, float] | str
ColorType = RGBColorType | RGBAColorType
HasCodeType = (
    types.ModuleType
    | type
    | types.MethodType
    | types.FunctionType
    | types.TracebackType
    | types.FrameType
    | types.CodeType
    | Callable[..., Any]
)



from __future__ import annotations

from collections.abc import Callable
import inspect
import os
import pathlib

from typing import Any, Self
from urllib import parse

from mknodes.basenodes import mkcontainer, mkfootnotes, mknode, processors
from mknodes.data import datatypes
from mknodes.pages import metadata, pagetemplate
from mknodes.utils import helpers, inspecthelpers, log, pathhelpers, reprhelpers


logger = log.get_logger(__name__)


class MkPage(mkcontainer.MkContainer):
    """A node container representing a Markdown page.

    A page contains a list of other Markdown nodes, has a virtual Markdown file
    associated, and can have metadata (added as header)
    """

    ICON = "fontawesome/solid/sheet-plastic"

    def __init__(
        self,
        title: str | None = None,
        *,
        hide: list[str] | str | None = None,
        search_boost: float | None = None,
        exclude_from_search: bool | None = None,
        icon: str | None = None,
        path: str | os.PathLike | None = None,
        status: datatypes.PageStatusStr | None = None,
        subtitle: str | None = None,
        description: str | None = None,
        template: str | pagetemplate.PageTemplate | None = None,
        inclusion_level: bool | None = None,
        tags: list[str] | None = None,
        edit_path: str | None = None,
        is_index: bool | None = None,
        is_homepage: bool | None = None,
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            path: Page path
            hide: Hide parts of the website ("toc", "nav", "path", "tags").
            search_boost: Factor to modify search ranking
            exclude_from_search: Whether to exclude this page from search listings
            icon: Optional page icon
            status: Page status
            title: Page title
            subtitle: Page subtitle
            description: Page description
            template: Page template (filename relative to `overrides` directory or
                       PageTemplate object)
            inclusion_level: Inclusion level of the page
            tags: tags to show above the main headline and within the search preview
            edit_path: Custom edit path for this page
            kwargs: Keyword arguments passed to parent
            is_index: Whether this page should be the index page.
            is_homepage: Whether this page should be the homepage.
        """
        super().__init__(**kwargs)
        self._path = str(path) if path else None
        self._edit_path = edit_path
        self._is_index = is_index
        self._is_homepage = is_homepage
        self.footnotes = mkfootnotes.MkFootNotes(parent=self)
        self.created_by: Callable | None = None
        self.metadata = metadata.Metadata(
            hide=hide,
            search_boost=search_boost,
            exclude_from_search=exclude_from_search,
            icon=icon,
            status=status,
            subtitle=subtitle,
            title=title,
            description=description,
            inclusion_level=inclusion_level,
            template=None,
            tags=tags,
        )
        self.template = template or pagetemplate.PageTemplate(
            parent=self,
            extends="main.html",
        )
        if frame := inspect.currentframe():
            self.metadata["created"] = inspecthelpers.get_stack_info(frame, level=2)
        logger.debug("Created %s, %r", type(self).__name__, self.resolved_file_path)

    def __repr__(self):
        kwargs = {k: v for k, v in self.metadata.items() if v is not None}
        return reprhelpers.get_repr(self, path=str(self.path), **kwargs)

    def is_index(self) -> bool:
        """Returns True if the page is the index page for the parent Nav."""
        return bool(self._is_index)

    @property
    def resolved_metadata(self) -> metadata.Metadata:
        """Return page metadata, complemented with the parent Nav metadata objects."""
        meta = metadata.Metadata()
        for nav in self.parent_navs:
            meta.update(nav.metadata)
        meta.update(self.metadata)
        return meta

    @property
    def path(self) -> str:
        """Return the last part of the page path."""
        if self._is_homepage:
            prefix = "../" * (len(self.parent_navs) - 1)
            return f"{prefix}index.md"
        if self._is_index:
            return "index.md"
        if self._path:
            path = self._path.removesuffix(".md") + ".md"
        else:
            path = f"{self.metadata.title}.md"
        return helpers.slugify(path)

    @path.setter
    def path(self, value: str | None):
        self._path = value

    @property
    def edit_url(self) -> str:
        base_url = parse.urljoin(self.ctx.metadata.repository_url, "edit/main/")
        return parse.urljoin(base_url, self._edit_path)

    @property
    def resolved_file_path(self) -> str:
        """Returns the resulting section/subsection/../filename.xyz path."""
        if self._is_homepage:
            return "index.md"
        path = "/".join(self.resolved_parts) + "/" + self.path
        return path.lstrip("/")

    @property
    def status(self) -> datatypes.PageStatusStr | str | None:
        """Return page status from metadata."""
        return self.metadata.status

    @status.setter
    def status(self, value: datatypes.PageStatusStr):
        self.metadata.status = value

    @property
    def title(self) -> str:
        """Return the page title if set, otherwise infer title from path."""
        return self.metadata.title or self.path.removesuffix(".md")

    @title.setter
    def title(self, value: str):
        self.metadata.title = value

    @property
    def subtitle(self) -> str | None:
        """Return subtitle from metadata."""
        return self.metadata.subtitle

    @subtitle.setter
    def subtitle(self, value: str):
        self.metadata.subtitle = value

    @property
    def icon(self) -> str | None:
        """Return page icon from metadata."""
        return self.metadata.icon

    @icon.setter
    def icon(self, value: str):
        self.metadata.icon = value

    @property
    def template(self) -> pagetemplate.PageTemplate:
        return self._template

    @template.setter
    def template(self, value: str | pagetemplate.PageTemplate | None):
        """Set the page template.

        If value is a string, use that string for metadata and clear the template object.
        If value is a PageTemplate, use that and put its name into metadata.

        Arguments:
            value: Page template to set.
        """
        if isinstance(value, pagetemplate.PageTemplate):
            self.metadata.template = value.filename
            self._template = value
        else:
            self.metadata.template = value
            self._template = None

    @classmethod
    def from_file(
        cls,
        path: str | os.PathLike,
        title: str | None = None,
        parent: mknode.MkNode | None = None,
        **kwargs: Any,
    ) -> Self:
        """Reads file content and creates an MkPage.

        Parses and reads header metadata.

        Arguments:
            path: Path to load file from, either a local path or a fsspec protocol path.
            title: Optional title to use
                   If None, title will be infered from metadata or filename
            parent: Optional parent for new page
            kwargs: Additional metadata for MkPage. Will override parsed metadata.
        """
        if "://" in (url := str(path)):
            file_content = pathhelpers.load_file_cached(url)
            path = f"{title}.md" if title else pathlib.Path(parse.urlsplit(url).path).name
            path = pathlib.Path(path)
        else:
            path = pathlib.Path(path)
            file_content = path.read_text()
        data, text = metadata.Metadata.parse(file_content)
        data.update(kwargs)
        page = cls(path=path.name, content=text, title=title, parent=parent)
        page.metadata.update(data)
        return page

    def get_processors(self) -> list[processors.TextProcessor]:
        """Override base MkNode processors."""
        return [
            processors.PrependMetadataProcessor(self.resolved_metadata),
            processors.FootNotesProcessor(self),
            processors.AnnotationProcessor(self),
        ]


if __name__ == "__main__":
    doc = MkPage.from_file(
        "https://raw.githubusercontent.com/mkdocs/mkdocs/master/docs/getting-started.md",
    )
    print(doc)
    # print(doc.children)



from __future__ import annotations

from typing import TYPE_CHECKING

from mknodes.mdlib import mdconverter
from mknodes.pages import templateblocks
from mknodes.utils import reprhelpers


if TYPE_CHECKING:
    import markdown

    import mknodes as mk


class PageTemplate:
    """A jinja template used to build a website from an MkPage."""

    def __init__(
        self,
        filename: str | None = None,
        extends: str | None = "base.html",
        parent: mk.MkPage | mk.MkNav | None = None,
    ):
        """Constructor.

        Arguments:
            filename: Optional filename for the Template (can get overridden during build)
            extends: Parent template path
            parent: The MkPage using this Page template.
        """
        self.filename = filename
        self.extends = f"{extends.rstrip('.html')}.html" if extends else None
        self.parent = parent

        # Common blocks
        self.title = templateblocks.TitleBlock()
        self.libs = templateblocks.LibsBlock()
        self.analytics = templateblocks.AnalyticsBlock()
        self.scripts = templateblocks.ScriptsBlock()
        self.site_meta = templateblocks.SiteMetaBlock()
        self.styles = templateblocks.StylesBlock()
        self.extra_head = templateblocks.ExtraHeadBlock()

        self.content_block = templateblocks.HtmlBlock("content", parent=parent)
        self.footer = templateblocks.HtmlBlock("footer", parent=parent)
        self.site_nav = templateblocks.HtmlBlock("site_nav", parent=parent)

        # MkDocs-Material
        self.tabs = templateblocks.HtmlBlock("tabs", parent=parent)
        self.outdated = templateblocks.HtmlBlock("outdated", parent=parent)
        self.hero = templateblocks.HtmlBlock("hero", parent=parent)
        self.announce = templateblocks.HtmlBlock("announce", parent=parent)

    def __bool__(self):
        return any(self.blocks)  # or self.extends

    def __hash__(self):
        return hash(self.build_html())

    @property
    def blocks(self) -> list[templateblocks.BaseBlock]:
        """A list containing all blocks of the PageTemplate."""
        return [
            self.title,
            self.content_block,
            self.tabs,
            self.announce,
            self.footer,
            self.libs,
            self.styles,
            self.outdated,
            self.hero,
            self.extra_head,
            self.site_nav,
            self.analytics,
            self.site_meta,
            self.scripts,
        ]

    def __repr__(self):
        return reprhelpers.get_repr(
            self,
            filename=self.filename,
            extends=self.extends,
            _filter_empty=True,
        )

    @property
    def announcement_bar(self):
        return self.announce.content

    @announcement_bar.setter
    def announcement_bar(self, value):
        import mknodes as mk

        if isinstance(value, mk.MkNode):
            value._ctx = self.parent.ctx
        self.announce.content = value

    @property
    def content(self):
        return self.content_block.content

    @content.setter
    def content(self, value):
        self.content_block.content = value

    def build_html(self, md: markdown.Markdown | None = None) -> str | None:
        """Convert given PageTemplate to HTML.

        Arguments:
            md: Markdown parser to use for HTML conversion.
        """
        md = md or mdconverter.MdConverter()
        blocks = [r'{% extends "' + self.extends + '" %}\n'] if self.extends else []
        blocks.extend(block.to_markdown(md) for block in self.blocks if block)
        return "\n".join(blocks) + "\n" if blocks else None


if __name__ == "__main__":
    import mknodes as mk

    md = mdconverter.MdConverter()
    template = PageTemplate(filename="main.html")
    template.announce.content = mk.MkAdmonition("test")
    html = template.build_html(md)
    print(html)



from __future__ import annotations

from .jinja.environment import Environment
from .jinja.nodeenvironment import NodeEnvironment

from .basenodes.mkblock import MkBlock
from .basenodes.mkhtmlblock import MkHtmlBlock
from .basenodes.mkcompactadmonition import MkCompactAdmonition
from .basenodes.mkadmonitionblock import MkAdmonitionBlock
from .basenodes.mkdetailsblock import MkDetailsBlock
from .basenodes.mkadmonition import MkAdmonition
from .basenodes.mknode import MkNode
from .basenodes.mkblockquote import MkBlockQuote
from .basenodes.mklink import MkLink
from .basenodes.mktext import MkText
from .basenodes.mkheader import MkHeader
from .basenodes.mkkeys import MkKeys
from .basenodes.mkgrid import MkGrid
from .basenodes.mkcontainer import MkContainer
from .basenodes.mkcode import MkCode
from .basenodes.mkcodeimage import MkCodeImage
from .basenodes._mkdocstrings import MkDocStrings
from .basenodes.mkclickdoc import MkClickDoc
from .basenodes.mkicon import MkIcon
from .basenodes.mkimage import MkImage
from .basenodes.mkmaterialbadge import MkMaterialBadge
from .basenodes.mkimagecompare import MkImageCompare
from .basenodes.mkimageslideshow import MkImageSlideshow
from .basenodes.mkbinaryimage import MkBinaryImage
from .basenodes.mklist import MkList
from .basenodes.mkdiagram import MkDiagram
from .basenodes.mktable import MkTable
from .basenodes.mkhtmltable import MkHtmlTable
from .basenodes.mktabcontainer import MkTabContainer
from .basenodes.mktabbedblocks import MkTabbedBlocks
from .basenodes.mktabbed import MkTabbed
from .basenodes.mksnippet import MkSnippet
from .basenodes.mkcritic import MkCritic
from .basenodes.mkannotations import MkAnnotations
from .basenodes.mkfootnotes import MkFootNotes
from .basenodes.mkiframe import MkIFrame
from .basenodes.mkprogressbar import MkProgressBar
from .basenodes.mkdefinitionlist import MkDefinitionList
from .basenodes.mkcard import MkCard
from .basenodes.mkshowcase import MkShowcase
from .basenodes.mkspeechbubble import MkSpeechBubble
from .basenodes.mktasklist import MkTaskList
from .basenodes.mktimeline import MkTimeline

from .pages.mkpage import MkPage
from .pages.mkclasspage import MkClassPage
from .pages.mkmodulepage import MkModulePage

from .navs.mknav import MkNav
from .navs.mkdoc import MkDoc

from .templatenodes.mkinclude import MkInclude
from .templatenodes.mkshields import MkShields
from .templatenodes.mkinstallguide import MkInstallGuide
from .templatenodes.mkchangelog import MkChangelog
from .templatenodes.mkcodeofconduct import MkCodeOfConduct
from .templatenodes.mkcommitconventions import MkCommitConventions
from .templatenodes.mkpullrequestguidelines import MkPullRequestGuidelines
from .templatenodes.mkdevenvsetup import MkDevEnvSetup
from .templatenodes.mkdevtools import MkDevTools
from .templatenodes.mkclassdiagram import MkClassDiagram
from .templatenodes.mkclasstable import MkClassTable
from .templatenodes.mkreprrawrendered import MkReprRawRendered
from .templatenodes.mkprettyprint import MkPrettyPrint
from .templatenodes.mkcallable import MkCallable
from .templatenodes.mktreeview import MkTreeView
from .templatenodes.mklicense import MkLicense
from .templatenodes.mkdependencytable import MkDependencyTable
from .templatenodes.mkcommandoutput import MkCommandOutput
from .templatenodes.mkbadge import MkBadge
from .templatenodes.mkmetadatabadges import MkMetadataBadges
from .templatenodes.mkmoduleoverview import MkModuleOverview
from .templatenodes.mkcommentedcode import MkCommentedCode
from .templatenodes.mkconfigsetting import MkConfigSetting
from .templatenodes.mkmoduletable import MkModuleTable
from .templatenodes.mkpluginflow import MkPluginFlow
from .templatenodes.mkargparsehelp import MkArgParseHelp
from .templatenodes.mkjinjatemplate import MkJinjaTemplate
from .templatenodes.mkjinjatext import MkJinjaText
from .templatenodes.mkpydeps import MkPyDeps
from .templatenodes.mkpipdeptree import MkPipDepTree

from .project import Project
from .pages.metadata import Metadata
from .pages.pagetemplate import PageTemplate
from .navs.mkdefaultwebsite import MkDefaultWebsite
from .navs.router import Router

from .theme.theme import Theme
from .theme.materialtheme import MaterialTheme

__all__ = [
    "Environment",
    "NodeEnvironment",
    "MkNode",
    "MkDefaultWebsite",
    "MkBlockQuote",
    "MkLink",
    "MkBlock",
    "MkHtmlBlock",
    "MkAdmonitionBlock",
    "MkCompactAdmonition",
    "MkDetailsBlock",
    "MkContainer",
    "MkNav",
    "MkDocStrings",
    "MkClickDoc",
    "MkText",
    "MkKeys",
    "MkGrid",
    "MkCode",
    "MkCodeImage",
    "MkMaterialBadge",
    "MkIcon",
    "MkImage",
    "MkImageCompare",
    "MkImageSlideshow",
    "MkHeader",
    "MkBinaryImage",
    "MkPage",
    "MkAdmonition",
    "MkDiagram",
    "MkClassDiagram",
    "MkTable",
    "MkHtmlTable",
    "MkClassTable",
    "MkList",
    "MkClassPage",
    "MkModulePage",
    "MkModuleTable",
    "MkPluginFlow",
    "MkArgParseHelp",
    "MkAnnotations",
    "MkFootNotes",
    "MkDoc",
    "MkTabContainer",
    "MkTabbedBlocks",
    "MkTabbed",
    "MkSnippet",
    "MkShields",
    "MkInclude",
    "MkCritic",
    "MkInstallGuide",
    "MkChangelog",
    "MkCodeOfConduct",
    "MkCommitConventions",
    "MkPullRequestGuidelines",
    "MkDevEnvSetup",
    "MkDevTools",
    "MkIFrame",
    "MkProgressBar",
    "MkDefinitionList",
    "MkCard",
    "MkShowcase",
    "MkReprRawRendered",
    "MkPrettyPrint",
    "MkCallable",
    "MkTreeView",
    "MkLicense",
    "MkDependencyTable",
    "MkCommandOutput",
    "MkBadge",
    "Project",
    "Metadata",
    "PageTemplate",
    "MkMetadataBadges",
    "MkModuleOverview",
    "MkCommentedCode",
    "MkConfigSetting",
    "MkSpeechBubble",
    "MkTaskList",
    "MkTimeline",
    "MkJinjaTemplate",
    "MkJinjaText",
    "MkPyDeps",
    "MkPipDepTree",
    "Router",
    "Theme",
    "MaterialTheme",
]

__version__ = "0.46.2"



from __future__ import annotations

from typing import Any

from mknodes.templatenodes import mkjinjatemplate
from mknodes.utils import log


logger = log.get_logger(__name__)


class MkPullRequestGuidelines(mkjinjatemplate.MkJinjaTemplate):
    """Node showing pull request guidelines."""

    ICON = "octicons/git-pull-request-24"

    def __init__(self, **kwargs: Any):
        """Constructor.

        Arguments:
            kwargs: Keyword arguments passed to parent
        """
        super().__init__(template="pullrequest_guidelines.jinja", **kwargs)

    @classmethod
    def create_example_page(cls, page):
        node = MkPullRequestGuidelines()
        page += node


if __name__ == "__main__":
    guideline = MkPullRequestGuidelines()
    print(guideline)
